# msgapi-sdk-php
Version: 2.0

## Installation
- Install **PHP 8.1** or later: [https://secure.php.net/manual/en/install.php](https://secure.php.net/manual/en/install.php). No further installation steps should be required. A 64-bit installation is recommended to prevent potential issues.

If you want to check whether your server meets the requirements and everything is configured properly, you can execute `./threema-msgapi-tool` without any parameters on the console or point your browser to the location where it is saved on your server.

## Build a PHAR file
Execute the build process:

```console
php ./bin/build.php
```

A ZIP file including the whole source and the PHAR file will be created in the `release` folder

## SDK usage
### Creating a connection

```php
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Receiver;

require_once('vendor/autoload.php');

//define your connection settings
$settings = new ConnectionSettings(
    '*THREEMA',
    'THISISMYSECRET'
);

//simple php file to store the public keys
$publicKeyStore = new Threema\MsgApi\PublicKeyStores\PhpFile('/path/to/my/keystore.php');

//create a connection
$connector = new Connection($settings, $publicKeyStore);
```

### Creating a connection with advanced options
**Attention:** These settings change internal values of the TLS connection. Choosing wrong settings can weaken the TLS connection or prevent connecting to the server. Use these options with care!

Each of the additional options shown below is optional. You can leave it out or use `null` to use the default value for this option.

```php
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Receiver;

require_once('vendor/autoload.php');

//define your connection settings
$settings = new ConnectionSettings(
    '*THREEMA',
    'THISISMYSECRET',
    null, //the host to be use, set to null for default (recommend)
    [
        'forceHttps' => true, //set to true to force HTTPS, default: false
        'tlsVersion' => '1.2', //set the version of TLS to be used, default: null
        'tlsCipher' => 'ECDHE-RSA-AES128-GCM-SHA256' //choose a cipher or a list of ciphers, default: null
    ]
);

//simple php file to store the public keys
$publicKeyStore = new Threema\MsgApi\PublicKeyStores\PhpFile('/path/to/my/keystore.php');

//create a connection
$connector = new Connection($settings, $publicKeyStore);
```

If you want to get a list of all ciphers, you can use have a look at the [SSLLabs scan](https://www.ssllabs.com/ssltest/analyze.html?d=msgapi.threema.ch) and at the list of all available [OpenSSL ciphers](https://www.openssl.org/docs/manmaster/apps/ciphers.html).

### Sending a text message to a Threema ID (Basic Mode)

```php
//create the connection
//(...)
//create a receiver
$receiver = new Receiver('ABCD1234', Receiver::TYPE_ID);

$result = $connector->sendSimple($receiver, "This is a Test Message");
if($result->isSuccess()) {
    echo 'new id created '.$result->getMessageId();
}
else {
    echo 'error '.$result->getErrorMessage();
}
```

### Sending a text message to a Threema ID (E2E Mode)

```php
//create the connection
//(...)

// Specify your own private key in hex below (without the "private:" prefix)
$senderPrivateKey = hex2bin("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef");

$e2eHelper = new \Threema\MsgApi\Helpers\E2EHelper($senderPrivateKey,$connector);
$result = $e2eHelper->sendTextMessage("TEST1234", "This is an end-to-end encrypted message");

if(true === $result->isSuccess()) {
    echo 'Message ID: '.$result->getMessageId() . "\n";
}
else {
    echo 'Error: '.$result->getErrorMessage() . "\n";
}
```

### Sending a file message to a Threema ID (E2E Mode)

```php
//create the connection
//(...)

// Specify your own private key in hex below (without the "private:" prefix)
$senderPrivateKey = hex2bin("0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef");
$filePath = "/path/to/my/file.pdf";

$e2eHelper = new E2EHelper($this->senderPrivateKey, $this->connector);
$result = $e2eHelper->sendFileMessage('TEST1234', $filePath, caption: '[PHP SDK] Test', renderingType: FileRenderingType::MEDIA);

if (true === $result->isSuccess()) {
  echo 'Message ID: ' . $result->getMessageId() . "\n";
} else {
  echo 'Error: ' . $result->getErrorMessage() . "\n";
}
```

## Console client usage
### Local operations (no network communication)
#### Encrypt

```console
./threema-msgapi-tool -e <privateKey> <publicKey>
```

Encrypt standard input using the given sender private key and recipient public key. two lines to standard output: first the nonce (hex), and then the box (hex).

#### Decrypt

```console
./threema-msgapi-tool -D <privateKey> <publicKey> <nonce>
```

Decrypt standard input using the given recipient private key and sender public key. The nonce must be given on the command line, and the box (hex) on standard input. Prints the decrypted message to standard output.

#### Hash Email Address

```console
./threema-msgapi-tool -h -e <email>
```

Hash an email address for identity lookup. Prints the hash in hex.

#### Hash Phone Number

```console
./threema-msgapi-tool -h -p <phoneNo>
```

Hash a phone number for identity lookup. Prints the hash in hex.

#### Generate Key Pair

```console
./threema-msgapi-tool -g <privateKeyFile> <publicKeyFile>
```

Generate a new key pair and write the private and public keys to the respective files (in hex).

#### Derive Public Key

```console
./threema-msgapi-tool -d <privateKey>
```

Derive the public key that corresponds with the given private key.

### Network operations
#### Send Simple Message

```console
./threema-msgapi-tool -s <threemaId> <from> <secret>
```

Send a message from standard input with server-side encryption to the given ID. 'from' is the API identity and 'secret'
is the API secret. Prints the message ID on success.

#### Send End-to-End Encrypted Text Message

```console
./threema-msgapi-tool -S <threemaId> <from> <secret> <privateKey>
```

Encrypt standard input and send the text message to the given ID. 'from' is the API identity and 'secret' is the API secret. Prints the message ID on success.

#### Send an End-to-End Encrypted File Message

```console
./threema-msgapi-tool -S -f <threemaId> <from> <secret> <privateKey> <file> <renderingType> <thumbnailFile> <caption>
```

Encrypt the file (with renderingType, thumbnail, caption if given) and send the message to the given ID. 'from' is the
API identity and 'secret' is the API secret. 'renderingType' must be either of 0 (file), 1 (media) or 2 (sticker).
Prints the message ID on success.

#### Send an End-to-End Encrypted Image Message (deprecated, use the File Message type instead)

```console
./threema-msgapi-tool -S -i <threemaId> <from> <secret> <privateKey> <imageFile>
```

Encrypt the image file and send the message to the given ID. 'from' is the API identity and 'secret' is the API secret.
Prints the message ID on success.

#### Send an End-to-End Encrypted Video Message (deprecated, use the File Message type instead)

```console
./threema-msgapi-tool -S -v <threemaId> <from> <secret> <privateKey> <duration> <videoFile> <thumbnailFile>
```

Encrypt the video and thumbnail file and send the message to the given ID. 'from' is the API
identity and 'secret' is the API secret. Prints the message ID on success.

#### Send an End-to-End Encrypted Audio Message (deprecated, use the File Message type instead)

```console
./threema-msgapi-tool -S -a <threemaId> <from> <secret> <privateKey> <duration> <audioFile>
```

Encrypt the audio file and send the message to the given ID. 'from' is the API
identity and 'secret' is the API secret. Prints the message ID on success.

#### Send an End-to-End Encrypted Location Message

```console
./threema-msgapi-tool -S -l <threemaId> <from> <secret> <privateKey> <latitude> <longitude> <accuracy> <poiName> <poiAddress>
```

Encrypt a location and send the message to the given ID. 'from' is the API
identity and 'secret' is the API secret. Prints the message ID on success.

#### Send an End-to-End Encrypted Ballot Create Message

```console
./threema-msgapi-tool -S -b -c <threemaId> <from> <secret> <privateKey> <ballotId> <ballotTitle> <ballotChoices> <ballotDisplayMode>
```

Send a ballot create message. Prints the message ID on success.

#### Send an End-to-End Encrypted Ballot Vote Message

```console
./threema-msgapi-tool -S -b -v <threemaId> <from> <secret> <privateKey> <ballotId> <ballotVotes>
```

Send a ballot vote message. Prints the message ID on success.

#### Send an End-to-End Encrypted Ballot Close Message

```console
./threema-msgapi-tool -S -b -d <threemaId> <from> <secret> <privateKey> <ballotId> <ballotTitle> <ballotChoices> <ballotResults> <ballotDisplayMode>
```

Send a ballot close message. Prints the message ID on success.

#### Send an End-to-End Encrypted Group Create Message

```console
./threema-msgapi-tool -G -c <threemaIds> <from> <secret> <privateKey> <groupId> <members>
```

Send a group create to the given IDs (members are comma-separated group members, no members = remove
group).

#### Send an End-to-End Encrypted Group Rename Message

```console
./threema-msgapi-tool -G -r <threemaIds> <from> <secret> <privateKey> <groupId> <subject>
```

Send a group rename to the given IDs.

#### Send an End-to-End Encrypted Set Group Photo Message

```console
./threema-msgapi-tool -G -p <threemaIds> <from> <secret> <privateKey> <groupId> <photo>
```

Send group avatar.

#### Send an End-to-End Encrypted Delete Group Photo Message

```console
./threema-msgapi-tool -G -d -p <threemaIds> <from> <secret> <privateKey> <groupId>
```

Delete group avatar.

#### Send an End-to-End Encrypted Group Ballot Create Message

```console
./threema-msgapi-tool -G -b -c <threemaIds> <from> <secret> <privateKey> <groupId> <ballotId> <ballotTitle> <ballotChoices> <ballotDisplayMode>
```

Send a group ballot create message. Prints the message IDs on success.

#### Send an End-to-End Encrypted Group Ballot Vote Message

```console
./threema-msgapi-tool -G -b -v <threemaIds> <from> <secret> <privateKey> <groupId> <ballotId> <ballotVotes>
```

Send a group ballot vote message. Prints the message IDs on success.

#### Send an End-to-End Encrypted Group Ballot Close Message

```console
./threema-msgapi-tool -G -b -d <threemaIds> <from> <secret> <privateKey> <groupId> <ballotId> <ballotTitle> <ballotChoices> <ballotResults> <ballotDisplayMode>
```

Send a group ballot close message. Prints the message IDs on success.

#### Send an End-to-End Encrypted Group Text Message

```console
./threema-msgapi-tool -G -S <threemaIds> <from> <secret> <privateKey> <groupId>
```

Encrypt standard input and send the group text message to the given IDs. 'from' is the API identity and
'secret' is the API secret. Prints the message IDs on success.

#### Send an End-to-End Encrypted Group File Message

```console
./threema-msgapi-tool -G -S -f <threemaIds> <from> <secret> <privateKey> <groupId> <file> <thumbnailFile> <caption>
```

Encrypt the file (and thumbnail if given) and send the message to the given ID. 'from' is the API
identity and 'secret' is the API secret. Prints the message ID on success.

#### Send an End-to-End Encrypted Group Location Message

```console
./threema-msgapi-tool -G -S -l <threemaIds> <from> <secret> <privateKey> <groupId> <latitude> <longitude> <accuracy> <poiName> <poiAddress>
```

Encrypt a group location and send it to the given IDs. 'from' is the API identity and 'secret'
is the API secret. Prints the message IDs on success.

#### ID-Lookup By Email Address

```console
./threema-msgapi-tool -l -e <email> <from> <secret>
```

Lookup the ID linked to the given email address (will be hashed locally).

#### ID-Lookup By Phone Number

```console
./threema-msgapi-tool -l -p <phoneNo> <from> <secret>
```

Lookup the ID linked to the given phone number (will be hashed locally).

#### Bulk ID-Lookup By Email Addresses and Phone Numbers

```console
./threema-msgapi-tool -l -B <from> <secret> <emailAddresses> <phoneNumbers>
```
Lookup IDs linked to the given email and phone address (will be hashed locally).

#### Fetch Public Key

```console
./threema-msgapi-tool -l -k <threemaId> <from> <secret>
```

Lookup the public key for the given ID.

#### Fetch Capabilities

```console
./threema-msgapi-tool -c <threemaId> <from> <secret>
```

Fetch the capabilities of a Threema ID.

#### Decrypt a Message and download the Files

```console
./threema-msgapi-tool -r <threemaId> <from> <secret> <privateKey> <messageId> <nonce> <outputFolder>
```

Decrypt a box (must be provided on stdin) message and download (if the message is an image or file message) the file(s) to the given <outputFolder> folder.

#### Remaining credits

```console
./threema-msgapi-tool -C <from> <secret>
```

Fetch remaining credits.
