<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use InvalidArgumentException;
use Threema\Console\Common;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Tools\ThreemaIdTool;
use Threema\MsgApi\Types\GroupId;

abstract class Base
{
	final public const argThreemaId = 'threemaId';
	final public const argThreemaIds = 'threemaIds';
	final public const argFrom = 'from';
	final public const argSecret = 'secret';
	final public const argPrivateKey = 'privateKey';
	final public const argPublicKey = 'publicKey';
	final public const argPrivateKeyFile = 'privateKeyFile';
	final public const argPublicKeyFile = 'publicKeyFile';
	final public const argNonce = 'nonce';
	final public const argGroupId = 'groupId';
	final public const argBallotId = 'ballotId';
	final public const argBallotTitle = 'ballotTitle';
	final public const argBallotChoices = 'ballotChoices';
	final public const argBallotVotes = 'ballotVotes';
	final public const argBallotResults = 'ballotResults';
	final public const argBallotDisplayMode = 'ballotDisplayMode';

	/** @var string[] */
	private array $arguments = [];

	public function __construct(private readonly string $subject, private readonly array $requiredArguments, private readonly string $description, private readonly array $optionalArguments = [])
	{
	}

	/**
	 * @throws \Threema\MsgApi\Core\Exception
	 */
	private function getIndexByPos(string $pos): int
	{
		$i = array_search($pos, $this->requiredArguments);
		if (false === $i) {
			$i = array_search($pos, $this->optionalArguments);
			if (false !== $i) {
				$i += count($this->requiredArguments);
			}
		}
		if (false === $i) {
			throw new Exception('argument ' . $pos . ' not found');
		}

		return $i;
	}

	/**
	 * @throws Exception
	 */
	private function getArgumentStringOrFileContent(string $pos): ?string
	{
		$content = $this->getArgument($pos);
		if (file_exists($content)) {
			$content = trim(file_get_contents($content));
		}

		return $content;
	}

	protected function readStdIn(): string
	{
		$f = fopen('php://stdin', 'r');

		$lines = [];
		while ($line = trim(fgets($f))) {
			if (strlen($line) == 0 || "\n" == $line) {
				continue;
			}

			$lines[] = $line;
		}

		return implode("\n", $lines);
	}

	public function getRequiredArgumentCount(): int
	{
		return null !== $this->requiredArguments ? count($this->requiredArguments) : 0;
	}

	public function getAllArgumentsCount(): int
	{
		return $this->getRequiredArgumentCount() + (null !== $this->optionalArguments ? count($this->optionalArguments) : 0);
	}

	/**
	 * @throws \Threema\MsgApi\Core\Exception
	 */
	public function getArgument(string $pos): ?string
	{
		$i = $this->getIndexByPos($pos);
		$required = $i < count($this->requiredArguments);

		if (false === array_key_exists($i, $this->arguments)) {
			if (true === $required) {
				throw new Exception('no argument at position ' . $i . ' (' . $pos . ')');
			}

			return null;
		}

		return $this->arguments[$i];
	}

	/**
	 * return a valid file path
	 *
	 * @throws \Threema\MsgApi\Core\Exception
	 */
	public function getArgumentFile(string $pos): ?string
	{
		$i = $this->getIndexByPos($pos);

		if (false === is_file($this->arguments[$i])) {
			throw new Exception('argument ' . $i . ' is not a file');
		}

		return $this->arguments[$i];
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentPrivateKey(string $pos): ?string
	{
		$content = Common::getPrivateKey($this->getArgumentStringOrFileContent($pos));
		if (null !== $content) {
			return hex2bin($content);
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentPublicKey(string $pos): ?string
	{
		$content = Common::getPublicKey($this->getArgumentStringOrFileContent($pos));
		if (null !== $content) {
			return hex2bin($content);
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentGroupId(string $pos): ?GroupId
	{
		$content = $this->getArgumentStringOrFileContent($pos);
		if (null !== $content && strlen($content) === 16) {
			return new GroupId(substr($content, 0, 8), substr($content, 8));
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentThreemaId(string $pos): ?string
	{
		$content = $this->getArgument($pos);
		if (null !== $content && strlen($content) == 8) {
			return strtoupper($content);
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentBallotId(string $pos): ?string
	{
		$content = $this->getArgument($pos);
		if (null !== $content && strlen($content) == 8) {
			return $content;
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentBallotTitle(string $pos): ?string
	{
		$content = $this->getArgument($pos);
		if (null !== $content && strlen(trim($content)) > 0) {
			return $content;
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentBallotChoices(string $pos): array
	{
		$content = $this->getArgument($pos);
		if (null !== $content && strlen($content) > 0) {
			return explode(',', $content);
		}

		return [];
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentBallotDisplayMode(string $pos): ?string
	{
		$content = $this->getArgument($pos);
		if (null !== $content && strlen(trim($content)) > 0) {
			return $content;
		}

		return null;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentBallotResults(string $pos): array
	{
		$content = $this->getArgument($pos);

		try {
			$results = @json_decode((string)$content, flags: JSON_THROW_ON_ERROR);
		} catch (\JsonException) {
			throw new InvalidArgumentException('invalid results delivered');
		}

		return $results;
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentBallotVotes(string $pos): array
	{
		$content = $this->getArgument($pos);
		if (null !== $content && strlen($content) > 0) {
			$votes = [];
			foreach (explode(',', $content) as $vote) {
				$votes[] = !(0 == $vote);
			}

			return $votes;
		}

		return [];
	}

	/**
	 * @throws Exception
	 */
	public function getArgumentThreemaIds(string $pos): ?array
	{
		$content = $this->getArgument($pos);

		if (null !== $content && mb_strlen($content) > 0) {
			$pieces = explode(',', strtoupper($content));
			$threemaIds = [];
			foreach ($pieces as $possibleThreemaId) {
				$possibleThreemaId = ThreemaIdTool::toThreemaId($possibleThreemaId);
				$threemaIds[$possibleThreemaId] = $possibleThreemaId;
			}

			if (!empty($threemaIds)) {
				return array_values($threemaIds);
			}
		}

		return null;
	}

	/**
	 * @throws \Threema\MsgApi\Core\Exception
	 */
	final public function run(array $arguments): void
	{
		$this->arguments = $arguments;

		$argCount = null !== $this->arguments ? count($this->arguments) : 0;

		if ($argCount < count($this->requiredArguments)) {
			throw new Exception('invalid count of arguments');
		}
		$this->doRun();
	}

	final public function help(bool $shellColors = true): string
	{
		$args = array_merge($this->requiredArguments, $this->optionalArguments);

		if (count($args) > 0) {
			$colorPrefix = '';
			$colorPostfix = '';
			if (true === $shellColors) {
				$colorPrefix = "\033[0;36m\033[40m";
				$colorPostfix = "\033[0m";
			}

			return  $colorPrefix . '<' . implode('> <', $args) . '>' . $colorPostfix;
		}

		return '';
	}

	final public function description(): string
	{
		return $this->description;
	}

	final public function subject(): string
	{
		return $this->subject;
	}

	abstract public function doRun();
}
