<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use Exception;
use InvalidArgumentException;
use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\Messages\GroupBallotCreateMessage;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Types\BallotId;
use Threema\MsgApi\Types\GroupId;
use Threema\MsgApi\Types\Voting\BallotChoice;
use Threema\MsgApi\Types\Voting\DisplayMode;
use Threema\MsgApi\Types\Voting\ResultsDisclosureType;
use Threema\MsgApi\Types\Voting\State;
use Threema\MsgApi\Types\Voting\VotingMode;

class GroupBallotClose extends Base
{
	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'send a group ballot close message',
			[
				self::argThreemaIds, self::argFrom, self::argSecret, self::argPrivateKey,
				self::argGroupId, self::argBallotId, self::argBallotTitle, self::argBallotChoices,
				self::argBallotResults, self::argBallotDisplayMode
			],
			'Send group ballot close message.'
		);
	}

	/**
	 * @throws \Threema\MsgApi\Core\Exception
	 * @throws \Threema\MsgApi\Exceptions\InvalidArgumentException
	 * @throws Exception
	 */
	public function doRun()
	{
		$threemaIds = $this->getArgumentThreemaIds(self::argThreemaIds);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$groupId = trim($this->getArgument(self::argGroupId));
		$ballotId = $this->getArgumentBallotId(self::argBallotId);
		$ballotTitle = $this->getArgumentBallotTitle(self::argBallotTitle);
		$ballotChoices = $this->getArgumentBallotChoices(self::argBallotChoices);
		$ballotResults = $this->getArgumentBallotResults(self::argBallotResults);
		$ballotDisplayMode = (int)$this->getArgumentBallotDisplayMode(self::argBallotDisplayMode);

		if (strlen($groupId) != 8) {
			throw new Exception('invalid group id ' . $groupId);
		}

		Common::required($threemaIds, $from, $secret, $privateKey);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);
		$helper = new E2EHelper($privateKey, $connector);
		$groupId = new GroupId($from, $groupId);

		$choices = [];
		foreach (array_values($ballotChoices) as $key => $choice) {
			if (is_string($choice) && !empty(trim($choice))) {
				$choices[] = new BallotChoice(
					$key,
					$choice,
					$key,
					array_values($ballotResults)[$key]
				);
			}
		}

		if (count($choices) < 2) {
			throw new InvalidArgumentException('please define at least two valid ballot choices');
		}

		if (count($choices) !== count($ballotResults)) {
			throw new InvalidArgumentException('results and choices should have the same length');
		}

		foreach ($ballotResults as $result) {
			if (2 !== (is_countable($result) ? count($result) : 0)) {
				throw new InvalidArgumentException('one result should belong to two threema ids');
			}
			foreach ($result as $key => $value) {
				if (0 !== $key && 1 !== $key) {
					throw new InvalidArgumentException('result index should belong to choice index');
				}
				if (0 !== $value && 1 !== $value) {
					throw new InvalidArgumentException('result value should only integer 0 or 1');
				}
			}
		}

		$ballotId = new BallotId($ballotId);
		$message = new GroupBallotCreateMessage(
			$groupId,
			$ballotId,
			$ballotTitle,
			State::CLOSED,
			VotingMode::MULTIPLE_CHOICE,
			ResultsDisclosureType::INTERMEDIATE,
			DisplayMode::from($ballotDisplayMode),
			$choices,
			array_merge([$from], $threemaIds)
		);

		// send ballot create
		$results = $helper->sendGroupBallotCreateMessages($message, $threemaIds);
		foreach ($results->getResults() as $result) {
			if ($result->isSuccess()) {
				Common::l('Group Ballot Close Message ID: ' . $result->getMessageId());
			} else {
				Common::e('Group Ballot Close Error: ' . $result->getErrorMessage());
			}
		}
	}
}
