<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use Exception;
use InvalidArgumentException;
use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\Messages\GroupBallotVoteMessage;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Types\BallotId;
use Threema\MsgApi\Types\GroupId;
use Threema\MsgApi\Types\Voting\VoteChoice;

class GroupBallotVote extends Base
{
	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'send a group ballot vote message',
			[self::argThreemaIds, self::argFrom, self::argSecret, self::argPrivateKey, self::argGroupId, self::argBallotId, self::argBallotVotes],
			'Send group ballot vote message.'
		);
	}

	/**
	 * @throws \Threema\MsgApi\Core\Exception
	 * @throws \Threema\MsgApi\Exceptions\InvalidArgumentException
	 * @throws Exception
	 */
	public function doRun()
	{
		$threemaIds = $this->getArgumentThreemaIds(self::argThreemaIds);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$groupId = trim($this->getArgument(self::argGroupId));
		$ballotId = $this->getArgumentBallotId(self::argBallotId);
		$ballotVotes = $this->getArgumentBallotVotes(self::argBallotVotes);

		if (strlen($groupId) != 8) {
			throw new Exception('invalid group id ' . $groupId);
		}

		Common::required($threemaIds, $from, $secret, $privateKey);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);
		$helper = new E2EHelper($privateKey, $connector);
		$groupId = new GroupId($from, $groupId);

		if (count($ballotVotes) < 2) {
			throw new InvalidArgumentException('please define at least two ballot votes');
		}

		$votes = [];
		foreach ($ballotVotes as $key => $vote) {
			$votes[] = new VoteChoice($key, 1 == $vote);
		}

		$ballotId = new BallotId($ballotId);

		// send ballot vote
		$results = $helper->sendGroupBallotVoteMessages(
			new GroupBallotVoteMessage(
				$groupId,
				$from,
				$ballotId,
				$votes
			),
			$threemaIds
		);
		foreach ($results->getResults() as $result) {
			if ($result->isSuccess()) {
				Common::l('Group Ballot Vote Message ID: ' . $result->getMessageId());
			} else {
				Common::e('Group Ballot Vote Error: ' . $result->getErrorMessage());
			}
		}
	}
}
