<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Types\FileRenderingType;

class GroupSendE2EFile extends Base
{
	final public const argFile = 'file';
	final public const argRenderingType = 'renderingType';
	final public const argThumbnail = 'thumbnailFile';
	final public const argCaption = 'caption';

	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'Send a End-to-End Encrypted File Message to a group',
			[self::argThreemaIds, self::argFrom, self::argSecret, self::argPrivateKey, self::argGroupId, self::argFile],
			'Encrypt the file (and thumbnail if given) and send the message to the given ID. \'from\' is the API identity and \'secret\' is the API secret. Prints the message ID on success.',
			[self::argRenderingType, self::argThumbnail, self::argCaption]
		);
	}

	/**
	 * @throws Exception
	 */
	public function doRun()
	{
		$threemaIds = $this->getArgumentThreemaIds(self::argThreemaIds);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$groupId = $this->getArgumentGroupId(self::argGroupId);
		$path = $this->getArgumentFile(self::argFile);
		$renderingType = $this->getArgument(self::argRenderingType);

		$thumbnailPath = $this->getArgument(self::argThumbnail);
		$caption = $this->getArgument(self::argCaption);

		Common::required($threemaIds, $from, $secret, $privateKey, $path);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);
		$helper = new E2EHelper($privateKey, $connector);

		$results = $helper->sendGroupFileMessages($groupId, $threemaIds, $path, $thumbnailPath, $caption, renderingType: FileRenderingType::from(intval($renderingType)));
		foreach ($results->getResults() as $threemaId => $result) {
			if ($result->isSuccess()) {
				Common::l("{$threemaId} Message ID : {$result->getMessageId()}");
			} else {
				Common::e("{$threemaId} Error: {$result->getErrorMessage()}");
			}
		}
	}
}
