<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Exceptions\BadMessageException;
use Threema\MsgApi\Exceptions\DecryptionFailedException;
use Threema\MsgApi\Exceptions\UnsupportedMessageTypeException;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\PublicKeyStore;

class ReceiveMessage extends Base
{
	final public const argOutputFolder = 'outputFolder';
	final public const argMessageId = 'messageId';

	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'Decrypt a Message and download the Files',
			[self::argThreemaId, self::argFrom, self::argSecret, self::argPrivateKey, self::argMessageId, self::argNonce],
			'Decrypt a box (must be provided on stdin) message and download (if the message is an image or file message) the file(s) to the given <' . self::argOutputFolder . '> folder',
			[self::argOutputFolder]
		);
	}

	/**
	 * @throws UnsupportedMessageTypeException
	 * @throws Exception
	 * @throws BadMessageException
	 * @throws DecryptionFailedException
	 */
	public function doRun()
	{
		$sendersThreemaId = $this->getArgumentThreemaId(self::argThreemaId);
		$id = $this->getArgumentThreemaId(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$nonce = hex2bin((string)$this->getArgument(self::argNonce));
		$messageId = $this->getArgument(self::argMessageId);
		$outputFolder = $this->getArgument(self::argOutputFolder);

		$box = hex2bin($this->readStdIn());

		Common::required($box, $id, $secret, $privateKey, $nonce);

		$settings = new ConnectionSettings(
			$id,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);
		$helper = new E2EHelper($privateKey, $connector);
		$message = $helper->receiveMessage(
			$sendersThreemaId,
			$messageId,
			$box,
			$nonce,
			$outputFolder
		);

		if (null === $message) {
			Common::e('invalid message');

			return;
		}

		if ($message->isSuccess()) {
			Common::l($message->getMessageId() . ' - ' . $message->getThreemaMessage());
			foreach ($message->getFiles() as $fileName => $filePath) {
				Common::l('   received file ' . $fileName . ' in ' . $filePath);
			}
		} else {
			Common::e('Error: ' . implode("\n", $message->getErrors()));
		}
	}
}
