<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use InvalidArgumentException;
use JsonException;
use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\Messages\BallotCreateMessage;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Types\BallotId;
use Threema\MsgApi\Types\Voting\BallotChoice;
use Threema\MsgApi\Types\Voting\DisplayMode;
use Threema\MsgApi\Types\Voting\ResultsDisclosureType;
use Threema\MsgApi\Types\Voting\State;
use Threema\MsgApi\Types\Voting\VotingMode;

class SendE2EBallotClose extends Base
{
	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'Send a End-to-End Encrypted Ballot Close Message',
			[
				self::argThreemaId, self::argFrom, self::argSecret, self::argPrivateKey,
				self::argBallotId, self::argBallotTitle, self::argBallotChoices, self::argBallotResults,
				self::argBallotDisplayMode
			],
			'Send an ballot close message. Prints the message ID on success.'
		);
	}

	/**
	 * @throws Exception
	 * @throws \Threema\MsgApi\Exceptions\InvalidArgumentException
	 * @throws JsonException
	 */
	public function doRun()
	{
		$threemaId = $this->getArgument(self::argThreemaId);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$ballotId = $this->getArgumentBallotId(self::argBallotId);
		$ballotTitle = $this->getArgumentBallotTitle(self::argBallotTitle);
		$ballotChoices = $this->getArgumentBallotChoices(self::argBallotChoices);
		$ballotResults = $this->getArgumentBallotResults(self::argBallotResults);
		$ballotDisplayMode = (int)$this->getArgumentBallotDisplayMode(self::argBallotDisplayMode);

		Common::required($threemaId, $from, $secret, $privateKey);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);

		$helper = new E2EHelper($privateKey, $connector);

		$choices = [];
		foreach (array_values($ballotChoices) as $key => $choice) {
			if (is_string($choice) && !empty(trim($choice))) {
				switch ($ballotDisplayMode) {
					case DisplayMode::LIST->value:
						$choices[] = new BallotChoice(
							$key,
							$choice,
							$key,
							array_values($ballotResults)[$key]
						);

						break;

					case DisplayMode::SUMMARY->value:
						$total = 0;
						foreach (array_values($ballotResults)[$key] as $value) {
							if (1 === $value) {
								$total++;
							}
						}
						$choices[] = new BallotChoice(
							$key,
							$choice,
							$key,
							totalVotes: $total
						);

						break;
					default:
						throw new InvalidArgumentException('unknown ballot display mode delivered');
				}
			}
		}

		if (count($choices) < 2) {
			throw new InvalidArgumentException('please define at least two valid ballot choices');
		}

		if (count($choices) !== count($ballotResults)) {
			throw new InvalidArgumentException('results and choices should have the same length');
		}

		foreach ($ballotResults as $result) {
			if (2 !== (is_countable($result) ? count($result) : 0)) {
				throw new InvalidArgumentException('one result should belong to two threema ids');
			}
			foreach ($result as $key => $value) {
				if (0 !== $key && 1 !== $key) {
					throw new InvalidArgumentException('result index should belong to choice index');
				}
				if (0 !== $value && 1 !== $value) {
					throw new InvalidArgumentException('result value should only integer 0 or 1');
				}
			}
		}

		$ballotId = new BallotId($ballotId);

		$message = new BallotCreateMessage(
			$ballotId,
			$ballotTitle,
			State::CLOSED,
			VotingMode::MULTIPLE_CHOICE,
			ResultsDisclosureType::INTERMEDIATE,
			DisplayMode::from($ballotDisplayMode),
			$choices,
			[$from, $threemaId]
		);

		// send ballot create
		$results = $helper->sendBallotCreateMessages($message, [$threemaId]);
		foreach ($results->getResults() as $result) {
			if ($result->isSuccess()) {
				Common::l('Ballot Close Message ID: ' . $result->getMessageId());
			} else {
				Common::e('Ballot Close Error: ' . $result->getErrorMessage());
			}
		}
	}
}
