<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use InvalidArgumentException;
use JsonException;
use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\Messages\BallotCreateMessage;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Types\BallotId;
use Threema\MsgApi\Types\Voting\BallotChoice;
use Threema\MsgApi\Types\Voting\DisplayMode;
use Threema\MsgApi\Types\Voting\ResultsDisclosureType;
use Threema\MsgApi\Types\Voting\State;
use Threema\MsgApi\Types\Voting\VotingMode;

class SendE2EBallotCreate extends Base
{
	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'Send a End-to-End Encrypted Ballot Create Message',
			[
				self::argThreemaId, self::argFrom, self::argSecret, self::argPrivateKey,
				self::argBallotId, self::argBallotTitle, self::argBallotChoices, self::argBallotDisplayMode
			],
			'Send an ballot create message. Prints the message ID on success.'
		);
	}

	/**
	 * @throws Exception
	 * @throws \Threema\MsgApi\Exceptions\InvalidArgumentException
	 * @throws JsonException
	 */
	public function doRun()
	{
		$threemaId = $this->getArgument(self::argThreemaId);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$ballotId = $this->getArgumentBallotId(self::argBallotId);
		$ballotTitle = $this->getArgumentBallotTitle(self::argBallotTitle);
		$ballotChoices = $this->getArgumentBallotChoices(self::argBallotChoices);
		$ballotDisplayMode = (int)$this->getArgumentBallotDisplayMode(self::argBallotDisplayMode);

		Common::required($threemaId, $from, $secret, $privateKey);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);

		$helper = new E2EHelper($privateKey, $connector);

		$choices = [];
		foreach (array_values($ballotChoices) as $key => $choice) {
			if (is_string($choice) && !empty(trim($choice))) {
				$choices[] = new BallotChoice(
					$key,
					$choice,
					$key,
					[],
					DisplayMode::SUMMARY === $ballotDisplayMode ? 0 : null
				);
			}
		}

		if (count($choices) < 2) {
			throw new InvalidArgumentException('please define at least two valid ballot choices');
		}

		$ballotId = new BallotId($ballotId);

		$message = new BallotCreateMessage(
			$ballotId,
			$ballotTitle,
			State::OPEN,
			VotingMode::MULTIPLE_CHOICE,
			ResultsDisclosureType::INTERMEDIATE,
			DisplayMode::from($ballotDisplayMode),
			$choices
		);

		// send ballot create
		$results = $helper->sendBallotCreateMessages($message, [$threemaId]);
		foreach ($results->getResults() as $result) {
			if ($result->isSuccess()) {
				Common::l('Ballot Create Message ID: ' . $result->getMessageId());
			} else {
				Common::e('Ballot Create Error: ' . $result->getErrorMessage());
			}
		}
	}
}
