<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use InvalidArgumentException;
use JsonException;
use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\Messages\BallotVoteMessage;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Types\BallotId;
use Threema\MsgApi\Types\Voting\VoteChoice;

class SendE2EBallotVote extends Base
{
	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'Send a End-to-End Encrypted Ballot Vote Message',
			[self::argThreemaId, self::argFrom, self::argSecret, self::argPrivateKey, self::argBallotId, self::argBallotVotes],
			'Send an ballot vote message. Prints the message ID on success.'
		);
	}

	/**
	 * @throws \Threema\MsgApi\Exceptions\InvalidArgumentException
	 * @throws Exception
	 * @throws JsonException
	 */
	public function doRun()
	{
		$threemaId = $this->getArgument(self::argThreemaId);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$ballotId = $this->getArgumentBallotId(self::argBallotId);
		$ballotVotes = $this->getArgumentBallotVotes(self::argBallotVotes);

		Common::required($threemaId, $from, $secret, $privateKey);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);

		$helper = new E2EHelper($privateKey, $connector);

		if (count($ballotVotes) < 2) {
			throw new InvalidArgumentException('please define at least two ballot votes');
		}

		$votes = [];
		foreach ($ballotVotes as $key => $vote) {
			$votes[] = new VoteChoice($key, 1 == $vote);
		}

		$ballotId = new BallotId($ballotId);
		// send ballot vote
		$results = $helper->sendBallotVoteMessages(
			new BallotVoteMessage(
				$from,
				$ballotId,
				$votes
			),
			[$threemaId]
		);
		foreach ($results->getResults() as $result) {
			if ($result->isSuccess()) {
				Common::l('Ballot Vote Message ID: ' . $result->getMessageId());
			} else {
				Common::e('Ballot Vote Error: ' . $result->getErrorMessage());
			}
		}
	}
}
