<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console\Command;

use JsonException;
use Threema\Console\Common;
use Threema\MsgApi\Connection;
use Threema\MsgApi\ConnectionSettings;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Helpers\E2EHelper;
use Threema\MsgApi\PublicKeyStore;

class SendE2ELocation extends Base
{
	final public const argLatitude = 'latitude';
	final public const argLongitude = 'longitude';
	final public const argAccuracy = 'accuracy';
	final public const argPOIName = 'poiName';
	final public const argPOIAddress = 'poiAddress';

	public function __construct(private readonly PublicKeyStore $publicKeyStore)
	{
		parent::__construct(
			'Send End-to-End Encrypted Location Message',
			[self::argThreemaId, self::argFrom, self::argSecret, self::argPrivateKey, self::argLatitude, self::argLongitude],
			'Encrypt a location and send the  message to the given ID. \'from\' is the API identity and \'secret\' is the API secret. Prints the message ID on success.',
			[self::argAccuracy, self::argPOIName, self::argPOIAddress]
		);
	}

	/**
	 * @throws Exception
	 * @throws JsonException
	 */
	public function doRun()
	{
		$threemaId = $this->getArgumentThreemaId(self::argThreemaId);
		$from = $this->getArgument(self::argFrom);
		$secret = $this->getArgument(self::argSecret);
		$privateKey = $this->getArgumentPrivateKey(self::argPrivateKey);
		$latitude = (float)$this->getArgument(self::argLatitude);
		$longitude = (float)$this->getArgument(self::argLongitude);
		$accuracy = (float)$this->getArgument(self::argAccuracy);
		$poiName = $this->getArgument(self::argPOIName);
		$poiAddress = $this->getArgument(self::argPOIAddress);

		Common::required($threemaId, $from, $secret, $privateKey, $latitude, $longitude);

		$settings = new ConnectionSettings(
			$from,
			$secret
		);

		$connector = new Connection($settings, $this->publicKeyStore);

		$helper = new E2EHelper($privateKey, $connector);
		$result = $helper->sendLocationMessage($threemaId, $latitude, $longitude, $accuracy, $poiName, $poiAddress);

		if ($result->isSuccess()) {
			Common::l('Message ID: ' . $result->getMessageId());
		} else {
			Common::e('Error: ' . $result->getErrorMessage());
		}
	}
}
