<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console;

use Threema\MsgApi\Constants;
use Threema\MsgApi\Core\Exception;

class Common
{
	/**
	 * output a string, wrap at 100 chars
	 *
	 * @param string $string string to output
	 * @param int $indent indent
	 */
	public static function l(string $string = '', int $indent = 0): void
	{
		$pad = str_repeat('  ', $indent);
		echo $pad . wordwrap($string, 100, "\n" . $pad) . "\n";
	}

	/**
	 * output a line
	 *
	 * @param string $string string to output
	 */
	public static function ln(string $string = ''): void
	{
		echo $string . "\n";
	}

	/**
	 * output a error message to the stderr
	 */
	public static function e(string $msg): void
	{
		$STDERR = fopen('php://stderr', 'w+');
		fwrite($STDERR, $msg . "\n");
	}

	/**
	 * check arguments for null, throws an exception on null or strlen == 0
	 *
	 * @params ...$things
	 * @throws \Threema\MsgApi\Core\Exception
	 */
	public static function required(): void
	{
		$argCount = func_num_args();
		for ($n = 0; $n < $argCount; $n++) {
			$o = func_get_arg($n);
			if (null === $o || (is_scalar($o) && strlen($o) == 0)) {
				throw new Exception('invalid data');
			}
		}
	}

	/**
	 * Append the prefix to the PublicKey @key
	 *
	 * @param string $key PublicKey in hex
	 */
	public static function convertPublicKey(string $key): ?string
	{
		return Constants::PUBLIC_KEY_PREFIX . $key;
	}

	/**
	 * Extract the PublicKey
	 *
	 * @param string|null $stringWithPrefix PublicKey in hex with or without the key-prefix
	 */
	public static function getPublicKey(?string $stringWithPrefix = null): ?string
	{
		if (null !== $stringWithPrefix) {
			if (str_starts_with($stringWithPrefix, Constants::PUBLIC_KEY_PREFIX)) {
				return substr($stringWithPrefix, strlen(Constants::PUBLIC_KEY_PREFIX));
			}
			return $stringWithPrefix;
		}

		return null;
	}

	/**
	 * Append the prefix to the PrivateKey @key
	 *
	 * @param string $key PrivateKey in hex
	 */
	public static function convertPrivateKey(string $key): ?string
	{
		return Constants::PRIVATE_KEY_PREFIX . $key;
	}

	/**
	 * Extract the PrivateKey
	 *
	 * @param string|null $stringWithPrefix PrivateKey in hex with or without the key-prefix (@Constants::PRIVATE_KEY_PREFIX)
	 */
	public static function getPrivateKey(?string $stringWithPrefix = null): ?string
	{
		if (null !== $stringWithPrefix) {
			if (str_starts_with($stringWithPrefix, Constants::PRIVATE_KEY_PREFIX)) {
				return substr($stringWithPrefix, strlen(Constants::PRIVATE_KEY_PREFIX));
			}

			return $stringWithPrefix;
		}

		return null;
	}
}
