<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\Console;

use Closure;
use Threema\Console\Command\Base;
use Threema\Console\Command\Capability;
use Threema\Console\Command\Credits;
use Threema\Console\Command\Decrypt;
use Threema\Console\Command\DerivePublicKey;
use Threema\Console\Command\Encrypt;
use Threema\Console\Command\GenerateKeyPair;
use Threema\Console\Command\GroupBallotClose;
use Threema\Console\Command\GroupBallotCreate;
use Threema\Console\Command\GroupBallotVote;
use Threema\Console\Command\GroupCreate;
use Threema\Console\Command\GroupDeletePhoto;
use Threema\Console\Command\GroupRename;
use Threema\Console\Command\GroupSendE2EFile;
use Threema\Console\Command\GroupSendE2ELocation;
use Threema\Console\Command\GroupSendE2EText;
use Threema\Console\Command\GroupSetPhoto;
use Threema\Console\Command\HashEmail;
use Threema\Console\Command\HashPhone;
use Threema\Console\Command\LookupIdBulk;
use Threema\Console\Command\LookupIdByEmail;
use Threema\Console\Command\LookupIdByPhoneNo;
use Threema\Console\Command\LookupPublicKeyById;
use Threema\Console\Command\ReceiveMessage;
use Threema\Console\Command\SendE2EAudio;
use Threema\Console\Command\SendE2EBallotClose;
use Threema\Console\Command\SendE2EBallotCreate;
use Threema\Console\Command\SendE2EBallotVote;
use Threema\Console\Command\SendE2EFile;
use Threema\Console\Command\SendE2EImage;
use Threema\Console\Command\SendE2ELocation;
use Threema\Console\Command\SendE2EText;
use Threema\Console\Command\SendE2EVideo;
use Threema\Console\Command\SendSimple;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\PublicKeyStore;
use Threema\MsgApi\Tools\CryptTool;

/**
 * Handling the console run stuff
 */
class Run
{
	/** @var \Threema\Console\Command\Base[] */
	private array $commands = [];

	private readonly string $scriptName;

	public function __construct(private array $arguments, private PublicKeyStore $publicKeyStore)
	{
		$this->scriptName = basename((string)array_shift($this->arguments));

		$this->registerSubject('Local operations (no network communication)');
		$this->register('-e', new Encrypt());
		$this->register('-D', new Decrypt());
		$this->register(['-h', '-e'], new HashEmail());
		$this->register(['-h', '-p'], new HashPhone());
		$this->register('-g', new GenerateKeyPair());
		$this->register('-d', new DerivePublicKey());

		$this->registerSubject('Network operations');
		//network operations
		$this->register('-s', new SendSimple($this->publicKeyStore));
		$this->register('-S', new SendE2EText($this->publicKeyStore));
		$this->register(['-S', '-i'], new SendE2EImage($this->publicKeyStore));
		$this->register(['-S', '-f'], new SendE2EFile($this->publicKeyStore));
		$this->register(['-S', '-v'], new SendE2EVideo($this->publicKeyStore));
		$this->register(['-S', '-a'], new SendE2EAudio($this->publicKeyStore));
		$this->register(['-S', '-l'], new SendE2ELocation($this->publicKeyStore));
		$this->register(['-S', '-b', '-c'], new SendE2EBallotCreate($this->publicKeyStore));
		$this->register(['-S', '-b', '-v'], new SendE2EBallotVote($this->publicKeyStore));
		$this->register(['-S', '-b', '-d'], new SendE2EBallotClose($this->publicKeyStore));
		$this->register(['-l', '-e'], new LookupIdByEmail($this->publicKeyStore));
		$this->register(['-l', '-p'], new LookupIdByPhoneNo($this->publicKeyStore));
		$this->register(['-l', '-B'], new LookupIdBulk($this->publicKeyStore));
		$this->register(['-l', '-k'], new LookupPublicKeyById($this->publicKeyStore));
		$this->register(['-G', '-c'], new GroupCreate($this->publicKeyStore));
		$this->register(['-G', '-r'], new GroupRename($this->publicKeyStore));
		$this->register(['-G', '-p'], new GroupSetPhoto($this->publicKeyStore));
		$this->register(['-G', '-b', '-c'], new GroupBallotCreate($this->publicKeyStore));
		$this->register(['-G', '-b', '-v'], new GroupBallotVote($this->publicKeyStore));
		$this->register(['-G', '-b', '-d'], new GroupBallotClose($this->publicKeyStore));
		$this->register(['-G', '-d', '-p'], new GroupDeletePhoto($this->publicKeyStore));
		$this->register(['-G', '-S'], new GroupSendE2EText($this->publicKeyStore));
		$this->register(['-G', '-S', '-f'], new GroupSendE2EFile($this->publicKeyStore));
		$this->register(['-G', '-S', '-l'], new GroupSendE2ELocation($this->publicKeyStore));
		$this->register(['-c'], new Capability($this->publicKeyStore));
		$this->register(['-r'], new ReceiveMessage($this->publicKeyStore));
		$this->register(['-C'], new Credits($this->publicKeyStore));
	}

	/**
	 * @throws Exception
	 */
	private function register(mixed $argumentKey, Base $command): static
	{
		if (null === $argumentKey) {
			return $this;
		}

		if (is_scalar($argumentKey)) {
			$argumentKey = [$argumentKey];
		}

		//check for existing commands with the same arguments
		foreach ($this->commands as $commandValues) {
			$ex = $commandValues[0];
			if (is_array($ex)) {
				if (count($ex) == (is_countable($argumentKey) ? count($argumentKey) : 0)
					&& count(array_diff($ex, $argumentKey)) == 0) {
					throw new Exception('arguments ' . implode($argumentKey) . ' already used');
				}
			}
		}
		$this->commands[] = [$argumentKey, $command];

		return $this;
	}

	private function registerSubject(string $name): void
	{
		$this->commands[] = $name;
	}

	private function help(): void
	{
		$defaultCryptTool = CryptTool::getInstance();

		Common::l();
		Common::l('Threema PHP MsgApi Tool');
		Common::l('Version: ' . MSGAPI_SDK_VERSION);
		Common::l('CryptTool: ' . $defaultCryptTool->getName() . ' ' . $defaultCryptTool->getDescription());
		Common::l(str_repeat('.', 40));
		Common::l();
		foreach ($this->commands as $data) {
			if (is_scalar($data)) {
				Common::l($data);
				Common::l(str_repeat('-', strlen((string)$data)));
			} else {
				list($key, $command) = $data;
				Common::ln($this->scriptName . ' ' . "\033[1;33m" . implode(' ', $key) . "\033[0m" . ' ' . $command->help());
				Common::l();
				// @noinspection PhpUndefinedMethodInspection
				Common::l($command->description(), 1);
			}
			Common::l();
		}
	}

	public function run(): void
	{
		$found = null;
		$argumentLength = 0;

		//find the correct command by arguments and arguments count
		foreach ($this->commands as $data) {
			if (is_scalar($data)) {
				continue;
			}

			list($keys, $command) = $data;
			if (array_slice($this->arguments, 0, is_countable($keys) ? count($keys) : 0) == $keys) {
				$argCount = count($this->arguments) - (is_countable($keys) ? count($keys) : 0);

				// @noinspection PhpUndefinedMethodInspection
				if ($argCount >= $command->getRequiredArgumentCount()
					&& $argCount <= $command->getAllArgumentsCount()) {
					$found = $command;
					$argumentLength = is_countable($keys) ? count($keys) : 0;

					break;
				}
			}
		}

		if ($argumentLength > 0) {
			array_splice($this->arguments, 0, $argumentLength);
		}

		if (null === $found) {
			$this->help();
		} else {
			try {
				$found->run($this->arguments);
			} catch (Exception $x) {
				Common::l();
				Common::e('ERROR: ' . $x->getMessage());
				Common::e($x::class);
				Common::l();
				die(1);
			}
		}
	}

	public function writeHelp(Closure $writer): void
	{
		if (null !== $writer) {
			foreach ($this->commands as $data) {
				if (is_scalar($data)) {
					$writer->__invoke($data, null, null, false);
				} else {
					list($key, $command) = $data;
					$writer->__invoke($command->subject(false), $this->scriptName . ' ' . implode(' ', $key) . ' ' . $command->help(false), $command->description(), true);
				}
			}
		}
	}
}
