<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\MsgApi\Commands\Results;

use Threema\MsgApi\Core\Exception;

class FetchPublicKeyBulkResult extends Result
{
	private array $publicKeys = [];

	protected function processResponse(?string $response): void
	{
		$result = @json_decode((string)$response, true);
		if (!empty($result)) {
			foreach ($result as $res) {
				$threemaId = $res['identity'] ?? null;
				$publicKey = $res['publicKey'] ?? null;
				if (!empty($threemaId) && !empty($publicKey)) {
					$this->publicKeys[$threemaId] = $publicKey;
				}
			}
		}
	}

	protected function getErrorMessageByErrorCode(int $httpCode): string
	{
		return match ($httpCode) {
			400 => 'JSON is invalid',
			401 => 'API identity or secret incorrect',
			413 => 'Too many identities have been specified in the request',
			500 => 'A temporary internal server error has occurred',
			default => 'Unknown error',
		};
	}

	/**
	 * @param array $publicKeys hex public keys indexed by threema ids
	 */
	public function addPublicKeys(array $publicKeys): void
	{
		// We have to loop, because array_merge will reset numeric array keys.
		foreach ($publicKeys as $threemaId => $publicKey) {
			$this->publicKeys[$threemaId] = $publicKey;
		}
	}

	/**
	 * @return array|string[] threema id indexed array of hex public keys
	 */
	public function getPublicKeys(): array
	{
		return $this->publicKeys;
	}

	/**
	 * @param bool $asBin get public key as binary instead of hexdeciaml string
	 * @throws Exception
	 * @return string hex of public key
	 */
	public function getPublicKey(string $threemaId, bool $asBin = true): string
	{
		if (false === array_key_exists($threemaId, $this->publicKeys)) {
			throw new Exception('Could not fetch public key');
		}

		return $asBin ? hex2bin((string)$this->publicKeys[$threemaId]) : $this->publicKeys[$threemaId];
	}
}
