<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\MsgApi;

use Closure;
use JsonException;
use Threema\MsgApi\Commands\Capability;
use Threema\MsgApi\Commands\CommandInterface;
use Threema\MsgApi\Commands\Credits;
use Threema\MsgApi\Commands\DownloadFile;
use Threema\MsgApi\Commands\FetchPublicKey;
use Threema\MsgApi\Commands\FetchPublicKeyBulk;
use Threema\MsgApi\Commands\JsonCommandInterface;
use Threema\MsgApi\Commands\LookupEmail;
use Threema\MsgApi\Commands\LookupIdBulk;
use Threema\MsgApi\Commands\LookupPhone;
use Threema\MsgApi\Commands\MultiPartCommandInterface;
use Threema\MsgApi\Commands\Results\CapabilityResult;
use Threema\MsgApi\Commands\Results\CreditsResult;
use Threema\MsgApi\Commands\Results\DownloadFileResult;
use Threema\MsgApi\Commands\Results\FetchPublicKeyBulkResult;
use Threema\MsgApi\Commands\Results\FetchPublicKeyResult;
use Threema\MsgApi\Commands\Results\LookupBulkIdResult;
use Threema\MsgApi\Commands\Results\LookupIdResult;
use Threema\MsgApi\Commands\Results\Result;
use Threema\MsgApi\Commands\Results\SendE2EBulkResult;
use Threema\MsgApi\Commands\Results\SendE2EResult;
use Threema\MsgApi\Commands\Results\SendSimpleResult;
use Threema\MsgApi\Commands\Results\UploadFileResult;
use Threema\MsgApi\Commands\SendE2E;
use Threema\MsgApi\Commands\SendE2EBulk;
use Threema\MsgApi\Commands\SendSimple;
use Threema\MsgApi\Commands\UploadFile;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Core\Url;

/**
 * Class Connection
 */
class Connection
{
	public function __construct(private readonly ConnectionSettings $setting, private readonly ?PublicKeyStore $publicKeyStore = null)
	{
	}

	private function createDefaultOptions(?Closure $progress = null): array
	{
		$options = [
			CURLOPT_RETURNTRANSFER => true
		];

		//no progress
		if (null !== $progress) {
			$options[CURLOPT_NOPROGRESS] = false;
			$options[CURLOPT_PROGRESSFUNCTION] = $progress;
		}

		//tls settings

		if (true === $this->setting->getTlsOption(ConnectionSettings::tlsOptionForceHttps, false)) {
			//limit allowed protocols to HTTPS
			$options[CURLOPT_PROTOCOLS] = CURLPROTO_HTTPS;
		}
		if ($tlsVersion = $this->setting->getTlsOption(ConnectionSettings::tlsOptionVersion)) {
			$options[CURLOPT_SSLVERSION] = match ($tlsVersion) {
				'1.0' => CURL_SSLVERSION_TLSv1_0,
				'1.1' => CURL_SSLVERSION_TLSv1_1,
				'1.2' => CURL_SSLVERSION_TLSv1_2,
				default => CURL_SSLVERSION_DEFAULT,
			};
		}
		if ($tlsCipher = $this->setting->getTlsOption(ConnectionSettings::tlsOptionCipher, null)) {
			if (true === is_string($tlsCipher)) {
				$options[CURLOPT_SSL_CIPHER_LIST] = $tlsCipher;
			}
		}

		return $options;
	}

	private function processRequestParams(array $params, ?array $more = null): array
	{
		$params['from'] = $this->setting->getThreemaId();
		$params['secret'] = $this->setting->getSecret();

		if (null !== $more) {
			$params = array_merge($params, $more);
		}

		return $params;
	}

	/**
	 * @param ?array $parameters
	 * @param ?Closure $result
	 * @throws Exception
	 * @return mixed
	 */
	private function call(string $path, array $curlOptions, ?array $parameters = null, ?Closure $result = null)
	{
		$fullPath = new Url('', $this->setting->getHost());
		$fullPath->addPath($path);

		if (null !== $parameters && count($parameters)) {
			foreach ($parameters as $key => $value) {
				$fullPath->setValue($key, $value);
			}
		}
		$session = curl_init($fullPath->getFullPath());
		curl_setopt_array($session, $curlOptions);

		$response = curl_exec($session);
		if (false === $response) {
			throw new Exception($path . ' ' . curl_error($session));
		}

		$httpCode = curl_getinfo($session, CURLINFO_HTTP_CODE);
		if (null === $result && 200 != $httpCode) {
			throw new Exception($httpCode);
		}

		if (null !== $result) {
			return $result->__invoke($httpCode, $response);
		}

		return $response;
	}

	/**
	 * @param ?Closure $progress
	 * @throws Exception
	 * @return Result
	 */
	protected function get(CommandInterface $command, ?Closure $progress = null)
	{
		$params = $this->processRequestParams($command->getParams());

		return $this->call(
			$command->getPath(),
			$this->createDefaultOptions($progress),
			$params,
			fn ($httpCode, $response) => $command->parseResult($httpCode, $response)
		);
	}

	/**
	 * @throws Exception
	 * @return Result
	 */
	protected function post(CommandInterface $command)
	{
		$options = $this->createDefaultOptions();
		$params = $this->processRequestParams($command->getParams());

		$options[CURLOPT_POST] = true;
		$options[CURLOPT_POSTFIELDS] = http_build_query($params);
		$options[CURLOPT_HTTPHEADER] = [
			'Content-Type: application/x-www-form-urlencoded'];

		return $this->call($command->getPath(), $options, null, fn ($httpCode, $response) => $command->parseResult($httpCode, $response));
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 * @return Result
	 */
	protected function postMultiPart(MultiPartCommandInterface $command, ?array $moreOptions = null)
	{
		$options = $this->createDefaultOptions();
		$params = $this->processRequestParams($command->getParams(), $moreOptions);

		$options[CURLOPT_POST] = true;
		$options[CURLOPT_HTTPHEADER] = ['Content-Type: multipart/form-data'];
		$options[CURLOPT_SAFE_UPLOAD] = true;
		$options[CURLOPT_POSTFIELDS] = array_merge([
			'blob' => $command->getData()
		], $command->getParams());

		return $this->call($command->getPath(), $options, $params, fn ($httpCode, $response) => $command->parseResult($httpCode, $response));
	}

	/**
	 * @throws \Exception
	 * @return Result
	 */
	protected function postJson(JsonCommandInterface $command)
	{
		$options = $this->createDefaultOptions();
		$params = $this->processRequestParams($command->getParams());

		$options[CURLOPT_POST] = true;
		$options[CURLOPT_HTTPHEADER] = ['Content-Type: application/json'];
		$options[CURLOPT_SAFE_UPLOAD] = true;

		try {
			$options[CURLOPT_POSTFIELDS] = json_encode($command->getData() ?? '', JSON_THROW_ON_ERROR);
		} catch (JsonException $e) {
			throw new \Exception('json encode operation has thrown an exception.');
		}

		return $this->call($command->getPath(), $options, $params, fn ($httpCode, $response) => $command->parseResult($httpCode, $response));
	}

	/**
	 * @throws Exception
	 */
	public function sendSimple(Receiver $receiver, string $text): Result|SendSimpleResult
	{
		$command = new SendSimple($receiver, $text);

		return $this->post($command);
	}

	/**
	 * @param ?array $options
	 * @param ?array $params
	 * @throws Exception
	 * @return Result|SendE2EBulkResult
	 */
	public function sendE2EBulk(
		array $threemaIds,
		array $nonces,
		array $boxes,
		?array $options = null,
		?array $params = null
	): SendE2EBulkResult {
		if (count($threemaIds) === 0
			|| count($threemaIds) !== count($nonces)
			|| count($threemaIds) !== count($boxes)) {
			throw new Exception('Invalid arguments');
		}
		$nonces = array_values($nonces);
		$boxes = array_values($boxes);

		$command = new SendE2EBulk($params);
		foreach (array_values($threemaIds) as $index => $threemaId) {
			$command->addCommand(new SendE2E(
				$threemaId,
				$nonces[$index],
				$boxes[$index],
				$options[$index] ?? null
			));
		}

		return $this->postJson($command);
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendE2E(string $threemaId, string $nonce, string $box, ?array $moreOptions = null): SendE2EResult
	{
		$command = new SendE2E($threemaId, $nonce, $box, $moreOptions);

		return $this->post($command);
	}

	/**
	 * @param ?array $moreOptions
	 * @param ?bool $persist
	 * @throws Exception
	 */
	public function uploadFile(string $encryptedFileData, ?array $moreOptions = null, ?bool $persist = false): UploadFileResult
	{
		$command = new UploadFile($encryptedFileData, $persist);

		return $this->postMultiPart($command, $moreOptions);
	}

	/**
	 * @param ?Closure $progress
	 * @throws Exception
	 */
	public function downloadFile(string $blobId, ?Closure $progress = null): DownloadFileResult
	{
		$command = new DownloadFile($blobId);

		return $this->get($command, $progress);
	}

	/**
	 * @throws Exception
	 */
	public function keyLookupByPhoneNumber(string $phoneNumber): LookupIdResult
	{
		$command = new LookupPhone($phoneNumber);

		return $this->get($command);
	}

	/**
	 * @throws Exception
	 */
	public function keyLookupByEmail(string $email): LookupIdResult
	{
		$command = new LookupEmail($email);

		return $this->get($command);
	}

	/**
	 * @param ?bool $hashed
	 * @throws \Exception
	 */
	public function keyLookupBulk(
		array $emails,
		array $phoneNumbers,
		?bool $hashed = true
	): LookupBulkIdResult {
		$command = new LookupIdBulk(
			$emails,
			$phoneNumbers,
			$hashed
		);

		return $this->postJson($command);
	}

	/**
	 * @param string $threemaId valid threema id (8 Chars)
	 * @throws Exception
	 */
	public function keyCapability(string $threemaId): CapabilityResult|Result
	{
		return $this->get(new Capability($threemaId));
	}

	/**
	 * @throws Exception
	 */
	public function credits(): CreditsResult|Result
	{
		return $this->get(new Credits());
	}

	/**
	 * @param mixed $threemaId
	 * @throws Exception
	 */
	public function fetchPublicKey($threemaId): FetchPublicKeyResult|Result
	{
		$publicKey = $this->publicKeyStore?->getPublicKey($threemaId);

		if (null === $publicKey) {
			$command = new FetchPublicKey($threemaId);
			$result = $this->get($command);
			if (false === $result->isSuccess()) {
				return $result;
			}
			$publicKey = $result->getRawResponse();

			$this->publicKeyStore?->setPublicKey($threemaId, $publicKey);
		}

		//create a key result
		return new FetchPublicKeyResult(200, $publicKey);
	}

	/**
	 * @throws \Exception
	 */
	public function fetchPublicKeyBulk(array $threemaIds): FetchPublicKeyBulkResult|Result
	{
		$publicKeys = [];

		if (null !== $this->publicKeyStore) {
			// fetch from keystore
			foreach ($threemaIds as $threemaId) {
				$publicKey = $this->publicKeyStore->getPublicKey($threemaId);
				if (null !== $publicKey) {
					$publicKeys[$threemaId] = $publicKey;
				}
			}
		}

		// get missing public keys
		$missingPublicKeys = array_diff($threemaIds, array_keys($publicKeys));

		// default response
		$response = new FetchPublicKeyBulkResult(200, null);

		// if we miss some public keys
		if (!empty($missingPublicKeys)) {
			$command = new FetchPublicKeyBulk($missingPublicKeys);
			$response = $this->postJson($command);
			if (false === $response->isSuccess()) {
				return $response;
			}

			if (null !== $this->publicKeyStore) {
				// add to key store
				foreach ($response->getPublicKeys() as $threemaId => $publicKey) {
					$this->publicKeyStore->setPublicKey($threemaId, $publicKey);
				}
			}
		}

		$response->addPublicKeys($publicKeys);

		return $response;
	}
}
