<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en
 */

namespace Threema\MsgApi\Core;

use Stringable;

class Url implements Stringable
{
	/** @var string[] */
	private array $values = [];

	public function __construct(private string $path, private readonly ?string $host = null)
	{
	}

	public function __toString(): string
	{
		return $this->getPath();
	}

	/**
	 * @return $this
	 */
	public function setValue(string $key, string $value): static
	{
		$this->values[$key] = $value;

		return $this;
	}

	/**
	 * Add a path to the current url
	 *
	 * @return $this
	 */
	public function addPath(string $path): static
	{
		while (str_ends_with($this->path, '/')) {
			$this->path = substr($this->path, 0, strlen($this->path) - 1);
		}

		$realPath = '';
		foreach (explode('/', $path) as $c => $pathPiece) {
			if ($c > 0) {
				$realPath .= '/';
			}
			$realPath .= urlencode($pathPiece);
		}
		while (str_starts_with($path, '/')) {
			$path = substr($path, 1);
		}

		$this->path .= '/' . $path;

		return $this;
	}

	public function getPath(): string
	{
		$p = $this->path;
		if (count($this->values) > 0) {
			$s = http_build_query($this->values);
			if (strlen($s) > 0) {
				$p .= '?' . $s;
			}
		}

		return $p;
	}

	public function getFullPath(): string
	{
		return $this->host . (str_starts_with($this->getPath(), '/') ? '' : '/') . $this->getPath();
	}

	public static function parametersToArray(string $urlParameter): array
	{
		$result = [];

		while (strlen($urlParameter) > 0) {
			// name
			$keyPosition = strpos($urlParameter, '=');
			$keyValue = substr($urlParameter, 0, $keyPosition);
			// value
			$valuePosition = strpos($urlParameter, '&') ?: strlen($urlParameter);
			$valueValue = substr($urlParameter, $keyPosition + 1, $valuePosition - $keyPosition - 1);

			// decoding the response
			$result[$keyValue] = urldecode($valueValue);
			$urlParameter = substr((string)$urlParameter, $valuePosition + 1, strlen((string)$urlParameter));
		}

		return $result;
	}

	public static function fromString($path): Url
	{
		$result = parse_url((string)$path);

		if (true === array_key_exists('scheme', $result)
			&& true === array_key_exists('host', $result)) {
			return new Url($result['path'], $result['scheme'] . '://' . $result['host']);
		}

		return new Url($path);
	}
}
