<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\MsgApi\Helpers;

use Closure;
use JsonException;
use Threema\MsgApi\Commands\Results\CapabilityResult;
use Threema\MsgApi\Commands\Results\DownloadFileResult;
use Threema\MsgApi\Commands\Results\SendE2EBulkResult;
use Threema\MsgApi\Commands\Results\SendE2EResult;
use Threema\MsgApi\Commands\Results\UploadFileResult;
use Threema\MsgApi\Connection;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Exceptions\BadMessageException;
use Threema\MsgApi\Exceptions\DecryptionFailedException;
use Threema\MsgApi\Exceptions\UnsupportedMessageTypeException;
use Threema\MsgApi\Messages\AudioMessage;
use Threema\MsgApi\Messages\BallotCreateMessage;
use Threema\MsgApi\Messages\BallotVoteMessage;
use Threema\MsgApi\Messages\DeliveryReceipt;
use Threema\MsgApi\Messages\FileMessage;
use Threema\MsgApi\Messages\GroupBallotCreateMessage;
use Threema\MsgApi\Messages\GroupBallotVoteMessage;
use Threema\MsgApi\Messages\GroupCreateMessage;
use Threema\MsgApi\Messages\GroupDeletePhoto;
use Threema\MsgApi\Messages\GroupDeliveryReceipt;
use Threema\MsgApi\Messages\GroupFileMessage;
use Threema\MsgApi\Messages\GroupLocationMessage;
use Threema\MsgApi\Messages\GroupRenameMessage;
use Threema\MsgApi\Messages\GroupRequestSyncMessage;
use Threema\MsgApi\Messages\GroupSetPhoto;
use Threema\MsgApi\Messages\GroupTextMessage;
use Threema\MsgApi\Messages\ImageMessage;
use Threema\MsgApi\Messages\LocationMessage;
use Threema\MsgApi\Messages\TextMessage;
use Threema\MsgApi\Messages\ThreemaMessage;
use Threema\MsgApi\Messages\VideoMessage;
use Threema\MsgApi\Tools\CryptTool;
use Threema\MsgApi\Tools\EncryptResult;
use Threema\MsgApi\Tools\FileAnalysisResult;
use Threema\MsgApi\Tools\FileAnalysisTool;
use Threema\MsgApi\Types\DeliveryReceiptType;
use Threema\MsgApi\Types\FileRenderingType;
use Threema\MsgApi\Types\GroupId;

class E2EHelper
{
	public function __construct(
		private readonly string $privateKey,
		private readonly Connection $connection,
		private ?CryptTool $cryptTool = null,
		private ?FileAnalysisTool $fileAnalysisTool = null
	) {
		if (null === $this->cryptTool) {
			$this->cryptTool = CryptTool::getInstance();
		}

		if (null === $this->fileAnalysisTool) {
			$this->fileAnalysisTool = new FileAnalysisTool();
		}
	}

	/**
	 * @param ?string $caption
	 * @param ?string $thumbnailPath
	 * @throws Exception
	 * @throws \Exception
	 * @return ProcessFileSendResult[] Result for all threemaIds
	 */
	private function processFileSending(
		array $threemaIds,
		string $filePath,
		ProcessFileSend $encryptFileMessageCallback,
		?string $caption = null,
		?string $thumbnailPath = null
	): array {
		//analyse the file
		$fileAnalysisResult = $this->fileAnalysisTool->analyse($filePath);

		if (null === $fileAnalysisResult) {
			throw new Exception('could not analyze the file (does it exist?)');
		}

		$thumbnailMimeType = null;
		if (strlen((string)$thumbnailPath) > 0) {
			if (true !== $this->fileAnalysisTool->isImage($thumbnailPath)) {
				throw new Exception('thumbnail is not a jpg or png' . $thumbnailPath);
			}
			$thumbnailMimeType = $this->fileAnalysisTool->analyse($thumbnailPath)->getMimeType();
		}

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		//encrypt the main file
		$encryptionResult = $this->cryptTool->encryptFile(
			$this->fileAnalysisTool->getContent($filePath)
		);

		$uploadResult = $this->connection->uploadFile($encryptionResult->getData(), [], count($threemaIds) > 1);

		if (null === $uploadResult) {
			throw new Exception('could not upload the file (no upload result)');
		}

		if (!$uploadResult->isSuccess()) {
			throw new Exception('could not upload the file (' . $uploadResult->getErrorCode() . ' ' . $uploadResult->getErrorMessage() . ') ' . $uploadResult->getRawResponse());
		}

		$thumbnailUploadResult = null;

		//encrypt the thumbnail file (if exists)
		if (strlen((string)$thumbnailPath) > 0) {
			//encrypt the main file
			$thumbnailEncryptionResult = $this->cryptTool->encryptFileThumbnail(
				$this->fileAnalysisTool->getContent($thumbnailPath),
				$encryptionResult->getKey()
			);
			$thumbnailUploadResult = $this->connection->uploadFile(
				$thumbnailEncryptionResult->getData()
			);

			if (!$thumbnailUploadResult->isSuccess()) {
				throw new Exception('could not upload the thumbnail file (' . $thumbnailUploadResult->getErrorCode() . ' ' . $thumbnailUploadResult->getErrorMessage() . ') ' . $thumbnailUploadResult->getRawResponse());
			}
		}

		$results = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->cryptTool->randomNonce();
			$box = $encryptFileMessageCallback->process(
				$this->cryptTool,
				$this->privateKey,
				$uploadResult,
				$encryptionResult,
				$fileAnalysisResult,
				$publicKeys->getPublicKey($threemaId),
				$nonce,
				$thumbnailUploadResult,
				$thumbnailMimeType,
				$caption
			);

			$results[$threemaId] = new ProcessFileSendResult($threemaId, $nonce, $box);
		}

		return $results;
	}

	/**
	 * @param string $blobId blob id as hex
	 * @param ?Closure $downloadMessage
	 * @throws Exception
	 */
	private function downloadFile(ThreemaMessage $message, string $blobId, ?Closure $downloadMessage = null): ?DownloadFileResult
	{
		if (null === $downloadMessage
			|| true === $downloadMessage->__invoke($message, $blobId)) {
			//make a download
			$result = $this->connection->downloadFile($blobId);
			if (false === $result->isSuccess()) {
				throw new Exception('could not download the file with blob id ' . $blobId);
			}

			return $result;
		}

		return null;
	}

	/**
	 * @param ?string $outputFolder
	 * @param ?Closure $downloadMessage
	 * @param ?string $ext
	 * @throws \Exception
	 */
	protected function processDownload(
		ThreemaMessage $message,
		string $blobId,
		string $nonce,
		Closure $decrypt,
		string $randomFilename,
		?string $outputFolder,
		?Closure $downloadMessage = null,
		?string $ext = null
	): ?string {
		if (null === $outputFolder || mb_strlen($outputFolder) === 0) {
			// No output folder = no download
			return null;
		}

		$result = $this->downloadFile($message, $blobId, $downloadMessage);
		if (null !== $result && true === $result->isSuccess()) {
			$decryptedContent = $decrypt->__invoke($result->getData(), $nonce);

			if (null === $decryptedContent) {
				throw new \Exception('decryption of blob failed');
			}

			if (null !== $ext) {
				$filePath = "{$outputFolder}/thumbnail-{$randomFilename}{$ext}";
			} else {
				$filePath = "{$outputFolder}/{$randomFilename}";
			}

			//save file
			$f = fopen($filePath, 'w+');
			fwrite($f, (string)$decryptedContent);
			fclose($f);

			return $decryptedContent;
		}

		return null;
	}

	protected function getCryptTool(): CryptTool
	{
		return $this->cryptTool;
	}

	protected function getConnection(): Connection
	{
		return $this->connection;
	}

	protected function getFileAnalysisTool(): ?FileAnalysisTool
	{
		return $this->fileAnalysisTool;
	}

	protected function sendE2E(
		string $threemaId,
		string $nonce,
		string $box,
		?array $moreOptions = null
	): SendE2EResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		return $this->connection->sendE2E($threemaId, $nonce, $box, $moreOptions);
	}

	protected function sendE2EBulk(
		array $threemaIds,
		array $nonces,
		array $boxes,
		?array $options = null,
		?array $params = null
	): SendE2EBulkResult {
		if (null === $options) {
			$options = [];
		}

		return $this->connection->sendE2EBulk($threemaIds, $nonces, $boxes, $options, $params);
	}

	/**
	 * Fetch a public key and check the capability of the threemaId
	 *
	 * @param ?Closure $capabilityCheck
	 * @throws Exception
	 * @return string Public key as binary
	 */
	protected function fetchPublicKeyAndCheckCapability(
		string $threemaId,
		?Closure $capabilityCheck = null
	): string {
		//fetch the public key
		$receiverPublicKey = $this->connection->fetchPublicKey($threemaId);

		if (null === $receiverPublicKey || !$receiverPublicKey->isSuccess()) {
			throw new Exception('Invalid threema id');
		}

		if (null !== $capabilityCheck) {
			//check capability
			$capability = $this->connection->keyCapability($threemaId);
			if (null === $capability || false === $capabilityCheck->__invoke($capability)) {
				throw new Exception('threema id does not have the capability');
			}
		}

		return hex2bin((string)$receiverPublicKey->getPublicKey());
	}

	/**
	 * @deprecated Use `E2EHelper::sendGroupCreates` instead
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupCreate(
		string $threemaId,
		string $groupId,
		array $members,
		?array $moreOptions = null
	): SendE2EResult {
		$res = $this->sendGroupCreates([$threemaId], $groupId, $members, $moreOptions);

		return current($res->getResults());
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupCreates(
		array $threemaIds,
		string $groupId,
		array $members,
		?array $moreOptions = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->getCryptTool()->padAndMakeBox(
				GroupCreateMessage::TYPE_CODE,
				$groupId . implode('', $members),
				$nonce,
				$this->getPrivateKey(),
				$publicKeys->getPublicKey($threemaId)
			);
			$options[] = array_merge($moreOptions[$key] ?? [], [
				'group' => true,
				'noPush' => true,
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @deprecated Use `E2EHelper::sendGroupPhotoMessages` instead
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupPhoto(
		string $threemaId,
		string $groupId,
		string $groupPhotoInJpgAsBinary,
		?array $moreOptions = null
	): SendE2EResult {
		$res = $this->sendGroupPhotoMessages([$threemaId], $groupId, $groupPhotoInJpgAsBinary, $moreOptions);

		return current($res->getResults());
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupPhotoMessages(
		array $threemaIds,
		string $groupId,
		string $groupPhotoInJpgAsBinary,
		?array $moreOptions = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$res = $this->getCryptTool()->encryptFile($groupPhotoInJpgAsBinary);
		$uploadRes = $this->getConnection()->uploadFile($res->getData());

		$message = $groupId;
//		echo PHP_EOL . 'size: ' . $res->getSize() . PHP_EOL;
		$message .= hex2bin($uploadRes->getBlobId());
		$message .= pack('V', $res->getSize());
		$message .= $res->getKey();

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->getCryptTool()->padAndMakeBox(
				GroupSetPhoto::TYPE_CODE,
				$message,
				$nonce,
				$this->getPrivateKey(),
				$publicKeys->getPublicKey($threemaId)
			);
			$options[$key] = array_merge($moreOptions[$key] ?? [], [
				'group' => true,
				'noPush' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @deprecated Use `E2EHelper::deleteGroupPhotos` instead
	 *
	 * Remove the group avatar photo
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function deleteGroupPhoto(string $threemaId, string $groupId, ?array $moreOptions = null): SendE2EResult
	{
		$res = $this->deleteGroupPhotos([$threemaId], $groupId, $moreOptions);

		return current($res->getResults());
	}

	/**
	 * Remove the group avatar photo
	 *
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function deleteGroupPhotos(array $threemaIds, string $groupId, ?array $moreOptions = null): SendE2EBulkResult
	{
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->getCryptTool()->padAndMakeBox(
				GroupDeletePhoto::TYPE_CODE,
				$groupId,
				$nonce,
				$this->getPrivateKey(),
				$publicKeys->getPublicKey($threemaId)
			);
			$options[$key] = array_merge($moreOptions[$key] ?? [], [
				'group' => true,
				'noPush' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @deprecated Use `E2EHelper::sendGroupRenames` instead
	 *
	 * Rename a group
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupRename(
		string $threemaId,
		string $groupId,
		string $groupName,
		?array $moreOptions = null
	): SendE2EResult {
		$res = $this->sendGroupRenames([$threemaId], $groupId, $groupName, $moreOptions);

		return current($res->getResults());
	}

	/**
	 * Rename a group
	 *
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupRenames(
		array $threemaIds,
		string $groupId,
		string $groupName,
		?array $moreOptions = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->getCryptTool()->padAndMakeBox(
				GroupRenameMessage::TYPE_CODE,
				$groupId . $groupName,
				$nonce,
				$this->getPrivateKey(),
				$publicKeys->getPublicKey($threemaId)
			);
			$options[$key] = array_merge($moreOptions[$key] ?? [], [
				'group' => true,
				'noPush' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	public function getPrivateKey(): string
	{
		return $this->privateKey;
	}

	/**
	 * Crypt a text message and send it to the threemaId
	 *
	 * @param string $text the text to be encrypted (max. 3500 bytes)
	 * @throws Exception
	 */
	public function sendTextMessage(string $threemaId, string $text): SendE2EResult
	{
		//random nonce first
		$nonce = $this->cryptTool->randomNonce();

		//fetch the public key
		$receiverPublicKey = $this->fetchPublicKeyAndCheckCapability($threemaId);

		$msg = new TextMessage(
			$text
		);

		//create a box
		$textMessage = $this->cryptTool->encryptTextMessage(
			$msg,
			$this->privateKey,
			$receiverPublicKey,
			$nonce
		);

		return $this->sendE2E($threemaId, $nonce, $textMessage);
	}

	/**
	 * @throws Exception
	 */
	public function sendTextMessages(array $threemaIds, string $text): SendE2EBulkResult
	{
		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$msg = new TextMessage(
			$text
		);

		$nonces = [];
		$boxes = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();
			$nonces[] = $nonce;

			$boxes[] = $this->cryptTool->encryptTextMessage(
				$msg,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes);
	}

	/**
	 * Crypt a location message and send it to the threemaId
	 *
	 * @param ?float $accuracy
	 * @param ?string $poiName
	 * @param ?string $poiAddress
	 * @throws Exception
	 */
	public function sendLocationMessage(
		string $threemaId,
		float $lat,
		float $lng,
		?float $accuracy = null,
		?string $poiName = null,
		?string $poiAddress = null
	): SendE2EResult {
		//random nonce first
		$nonce = $this->cryptTool->randomNonce();

		//fetch the public key
		$receiverPublicKey = $this->fetchPublicKeyAndCheckCapability($threemaId);

		$msg = new LocationMessage(
			$lat,
			$lng,
			$poiAddress,
			$accuracy,
			$poiName
		);

		//create a box
		$textMessage = $this->cryptTool->encryptLocationMessage(
			$msg,
			$this->privateKey,
			$receiverPublicKey,
			$nonce
		);

		return $this->sendE2E($threemaId, $nonce, $textMessage);
	}

	/**
	 * Crypt a group text message and send it to the threemaId
	 *
	 * @deprecated Use `E2EHelper::sendGroupTextMessages` instead
	 * @param ?array $moreOptions
	 * @throws Exception|JsonException
	 */
	public function sendGroupTextMessage(
		GroupId $groupId,
		string $threemaId,
		string $text,
		?array $moreOptions = null
	): SendE2EResult {
		$result = $this->sendGroupTextMessages($groupId, [$threemaId], $text, $moreOptions);

		return current($result->getResults());
	}

	/**
	 * Crypt a group text message and send it to the threemaId
	 *
	 * @param ?array $moreOptions
	 * @throws Exception
	 * @throws \Exception
	 */
	public function sendGroupTextMessages(
		GroupId $groupId,
		array $threemaIds,
		string $text,
		?array $moreOptions = null
	): SendE2EBulkResult {
		$options = $moreOptions;
		if (null === $options) {
			$options = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$msg = new GroupTextMessage(
			$groupId,
			$text
		);

		$options['group'] = true;

		$nonces = [];
		$boxes = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptGroupTextMessage(
				$msg,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @deprecated Use `E2EHelper::sendGroupLocationMessages` instead
	 * Crypt a group location message and send it to the threemaId
	 * @param ?float $accuracy
	 * @param ?string $poiName
	 * @param ?string $poiAddress
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupLocationMessage(
		GroupId $groupId,
		string $threemaId,
		float $lat,
		float $lng,
		?float $accuracy = null,
		?string $poiName = null,
		?string $poiAddress = null,
		?array $moreOptions = null
	): SendE2EResult {
		$result = $this->sendGroupLocationMessages($groupId, [$threemaId], $lat, $lng, $accuracy, $poiName, $poiAddress, $moreOptions);

		return current($result->getResults());
	}

	/**
	 * Crypt a group location message and send it to the threemaId
	 *
	 * @param ?float $accuracy
	 * @param ?string $poiName
	 * @param ?string $poiAddress
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupLocationMessages(
		GroupId $groupId,
		array $threemaIds,
		float $lat,
		float $lng,
		?float $accuracy = null,
		?string $poiName = null,
		?string $poiAddress = null,
		?array $moreOptions = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$msg = new GroupLocationMessage(
			$groupId,
			$lat,
			$lng,
			$poiAddress,
			$accuracy,
			$poiName
		);

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach (array_unique($threemaIds) as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptGroupLocationMessage(
				$msg,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
			$options[$key] = array_merge($moreOptions[$key] ?? [], [
				'group' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * Crypt a file (and thumbnail if given), upload the blob and send it to the given threemaId
	 *
	 * @deprecated Use `E2EHelper::sendGroupFileMessages` instead
	 * @param ?string $thumbnailPath
	 * @param ?string $caption
	 * @param ?array $moreOptions
	 * @param ?string $fileName
	 * @param ?FileRenderingType $renderingType
	 * @throws Exception
	 */
	public function sendGroupFileMessage(
		GroupId $groupId,
		string $threemaId,
		string $filePath,
		?string $thumbnailPath = null,
		?string $caption = null,
		?array $moreOptions = null,
		?string $fileName = null,
		?FileRenderingType $renderingType = null
	): SendE2EResult {
		$results = $this->sendGroupFileMessages($groupId, [$threemaId], $filePath, $thumbnailPath, $caption, $moreOptions, $fileName, $renderingType);

		return current($results->getResults());
	}

	/**
	 * Crypt a file (and thumbnail if given), upload the blob and send it to the given threemaIds
	 *
	 * @param ?string $thumbnailPath
	 * @param ?string $caption
	 * @param ?array $moreOptions
	 * @param ?string $fileName
	 * @param ?FileRenderingType $renderingType
	 * @param ?string $correlationId
	 * @param ?array $metadata
	 * @throws Exception
	 */
	public function sendGroupFileMessages(
		GroupId $groupId,
		array $threemaIds,
		string $filePath,
		?string $thumbnailPath = null,
		?string $caption = null,
		?array $moreOptions = null,
		?string $fileName = null,
		?FileRenderingType $renderingType = FileRenderingType::FILE,
		?string $correlationId = null,
		?array $metadata = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		$fileMessages = $this->processFileSending($threemaIds, $filePath, new class ($groupId, $fileName, $renderingType, $correlationId, $metadata) implements ProcessFileSend {
			public function __construct(private readonly GroupId $groupId, private readonly ?string $fileName = null, private readonly ?FileRenderingType $renderingType = null, private readonly ?string $correlationId = null, private readonly ?array $metadata = null)
			{
			}

			public function process(
				CryptTool $cryptTool,
				string $privateKey,
				UploadFileResult $uploadResult,
				EncryptResult $encryptionResult,
				FileAnalysisResult $fileAnalysisResult,
				string $receiverPublicKey,
				string $nonce,
				?UploadFileResult $thumbnailUploadResult,
				?string $thumbnailMimeType,
				?string $caption
			): string {
				$msg = new GroupFileMessage(
					$this->groupId,
					$fileAnalysisResult->getMimeType(),
					$fileAnalysisResult->getSize(),
					$uploadResult->getBlobId(),
					$encryptionResult->getKey(),
					$thumbnailUploadResult?->getBlobId(),
					$thumbnailMimeType,
					$caption,
					$this->renderingType,
					$this->correlationId,
					$this->metadata,
					$this->fileName ?? $fileAnalysisResult->getFileName()
				);

				return $cryptTool->encryptGroupFileMessage(
					$msg,
					$privateKey,
					$receiverPublicKey,
					$nonce
				);
			}
		}, $caption, $thumbnailPath);

		$nonces = [];
		$boxes = [];
		$realThreemaIds = [];
		$options = [];
		$key = 0;
		foreach ($fileMessages as $threemaId => $result) {
			$realThreemaIds[] = $threemaId;
			$nonces[] = $result->getNonce();
			$boxes[] = $result->getBox();
			$options[] = array_merge($moreOptions[$key] ?? [], [
				'group' => true,
				'persist' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($realThreemaIds, $nonces, $boxes, $options);
	}

	/* ============================
	 * Status update message types
	 * ============================ */

	/**
	 * Encrypt a group request sync message and send it to the given group.
	 *
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupRequestSyncMessage(
		GroupId $groupId,
		array $threemaIds,
		?array $moreOptions = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$msg = new GroupRequestSyncMessage(
			$groupId
		);

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach (array_unique($threemaIds) as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptGroupRequestSyncMessage(
				$msg,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
			$options[$key] = array_merge($moreOptions[$key] ?? [], [
				'group' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * Encrypt a delivery receipt and send it to the original sender.
	 *
	 * @throws Exception
	 */
	public function sendDeliveryReceipt(
		string $threemaId,
		array $ackedMessageIds,
		DeliveryReceiptType $receiptType
	): SendE2EResult {
		//random nonce first
		$nonce = $this->cryptTool->randomNonce();

		//fetch the public key
		$receiverPublicKey = $this->fetchPublicKeyAndCheckCapability($threemaId);

		$msg = new DeliveryReceipt(
			$receiptType,
			$ackedMessageIds
		);

		$message = $this->cryptTool->encryptDeliveryReceipt(
			$msg,
			$this->privateKey,
			$receiverPublicKey,
			$nonce
		);

		return $this->sendE2E($threemaId, $nonce, $message);
	}

	/**
	 * Encrypt a group delivery receipt and send it to all group members.
	 *
	 * @throws Exception
	 */
	public function sendGroupDeliveryReceipt(
		GroupId $groupId,
		array $threemaIds,
		array $ackedMessageIds,
		DeliveryReceiptType $receiptType,
		?array $moreOptions = null
	): SendE2EBulkResult {
		if (null === $moreOptions) {
			$moreOptions = [];
		}

		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$msg = new GroupDeliveryReceipt(
			$groupId,
			$receiptType,
			$ackedMessageIds
		);

		$nonces = [];
		$boxes = [];
		$options = [];
		$key = 0;
		foreach (array_unique($threemaIds) as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptGroupDeliveryReceipt(
				$msg,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
			$options[$key] = array_merge($moreOptions[$key] ?? [], [
				'group' => true
			]);
			$key++;
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/* ========================
	 * (Deprecated) Legacy types
	 * ======================== */

	/**
	 * Crypt a image file, upload the blob and send the image message to the threemaId
	 * NOTE: This message type is being phased out
	 *
	 * @throws Exception
	 * @deprecated
	 */
	public function sendImageMessage(string $threemaId, string $imagePath): SendE2EResult
	{
		//analyse the file
		$fileAnalyzeResult = $this->fileAnalysisTool->analyse($imagePath);

		if (null === $fileAnalyzeResult) {
			throw new Exception('could not analyze the file');
		}

		if (true !== $this->fileAnalysisTool->isImage($imagePath)) {
			throw new Exception('file is not a jpg or png');
		}

		//fetch the public key
		$receiverPublicKey = $this->fetchPublicKeyAndCheckCapability($threemaId, fn (CapabilityResult $capabilityResult) => true === $capabilityResult->canImage());

		//encrypt the image file
		$encryptionResult = $this->cryptTool->encryptImage(
			$this->fileAnalysisTool->getContent($imagePath),
			$this->privateKey,
			$receiverPublicKey
		);

		$uploadResult = $this->connection->uploadFile($encryptionResult->getData());

		if (null == $uploadResult || !$uploadResult->isSuccess()) {
			throw new Exception('could not upload the image (' . $uploadResult->getErrorCode() . ' ' . $uploadResult->getErrorMessage() . ') ' . $uploadResult->getRawResponse());
		}

		$nonce = $this->cryptTool->randomNonce();

		//create an image message box
		$imageMessage = $this->cryptTool->encryptImageMessage(
			$uploadResult,
			$encryptionResult,
			$this->privateKey,
			$receiverPublicKey,
			$nonce
		);

		return $this->sendE2E($threemaId, $nonce, $imageMessage);
	}

	/**
	 * Crypt a video file (and thumbnail if given), upload the blob and send it to the given threemaId
	 * NOTE: This message type is being phased out
	 *
	 * @deprecated
	 * @param ?string $videoThumbnailPath
	 * @throws Exception
	 */
	public function sendVideoMessage(
		string $threemaId,
		int $duration,
		string $videoPath,
		?string $videoThumbnailPath = null
	): SendE2EResult {
		//analyse the files
		$fileAnalyzeResult = $this->fileAnalysisTool->analyse($videoPath);

		if (null === $fileAnalyzeResult) {
			throw new Exception('could not analyze the video file');
		}

		//analyse the files
		if (true !== $this->fileAnalysisTool->isImage($videoThumbnailPath)) {
			throw new Exception('thumbnail is not a jpg or png');
		}

		if (false === is_numeric($duration)) {
			throw new Exception('duration must be a integer');
		}

		//fetch the public key
		$receiverPublicKey = $this->fetchPublicKeyAndCheckCapability($threemaId, fn (CapabilityResult $capabilityResult) => true === $capabilityResult->canVideo());

		//encrypt the main file
		$encryptionResult = $this->cryptTool->encryptFile(
			$this->fileAnalysisTool->getContent($videoPath)
		);
		$uploadResult = $this->connection->uploadFile($encryptionResult->getData());

		if (null === $uploadResult) {
			throw new Exception('upload failed (no upload result)');
		}

		if (!$uploadResult->isSuccess()) {
			throw new Exception('could not upload the file (' . $uploadResult->getErrorCode() . ' ' . $uploadResult->getErrorMessage() . ') ' . $uploadResult->getRawResponse());
		}

		//encrypt the thumbnail file
		$thumbnailEncryptionResult = $this->cryptTool->encryptFileThumbnail(
			$this->fileAnalysisTool->getContent($videoThumbnailPath),
			$encryptionResult->getKey()
		);

		$thumbnailUploadResult = $this->connection->uploadFile(
			$thumbnailEncryptionResult->getData()
		);

		if (!$thumbnailUploadResult->isSuccess()) {
			throw new Exception('could not upload the thumbnail file (' . $thumbnailUploadResult->getErrorCode() . ' ' . $thumbnailUploadResult->getErrorMessage() . ') ' . $thumbnailUploadResult->getRawResponse());
		}

		$nonce = $this->cryptTool->randomNonce();

		//create a file message box
		$fileMessage = $this->cryptTool->encryptVideoMessage(
			$duration,
			$uploadResult,
			$thumbnailUploadResult,
			$encryptionResult,
			$thumbnailEncryptionResult,
			$this->privateKey,
			$receiverPublicKey,
			$nonce
		);

		return $this->sendE2E($threemaId, $nonce, $fileMessage);
	}

	/**
	 * Crypt a audio file, upload the blob and send it to the given threemaId
	 * NOTE: This message type is being phased out
	 *
	 * @throws Exception
	 * @deprecated
	 */
	public function sendAudioMessage(
		string $threemaId,
		int $duration,
		string $audioPath
	): SendE2EResult {
		//analyse the files
		$fileAnalyzeResult = $this->fileAnalysisTool->analyse($audioPath);

		if (null === $fileAnalyzeResult) {
			throw new Exception('could not analyze the audio file');
		}

		if (false === is_numeric($duration)) {
			throw new Exception('duration must be a integer');
		}

		//fetch the public key
		$receiverPublicKey = $this->fetchPublicKeyAndCheckCapability($threemaId, fn (CapabilityResult $capabilityResult) => true === $capabilityResult->canAudio());

		//encrypt the main file
		$encryptionResult = $this->cryptTool->encryptFile(
			$this->fileAnalysisTool->getContent($audioPath)
		);
		$uploadResult = $this->connection->uploadFile($encryptionResult->getData());

		if (null == $uploadResult || !$uploadResult->isSuccess()) {
			throw new Exception('could not upload the file (' . $uploadResult->getErrorCode() . ' ' . $uploadResult->getErrorMessage() . ') ' . $uploadResult->getRawResponse());
		}

		$nonce = $this->cryptTool->randomNonce();

		//create a file message box
		$fileMessage = $this->cryptTool->encryptAudioMessage(
			$duration,
			$uploadResult,
			$encryptionResult,
			$this->privateKey,
			$receiverPublicKey,
			$nonce
		);

		return $this->sendE2E($threemaId, $nonce, $fileMessage);
	}

	/**
	 * Crypt a file (and thumbnail if given), upload the blob and send it to the given threemaId
	 *
	 * @param ?string $thumbnailPath
	 * @param ?string $caption
	 * @param ?string $fileName
	 * @param ?FileRenderingType $renderingType
	 * @param ?string $correlationId
	 * @param ?array $metadata
	 * @throws Exception
	 */
	public function sendFileMessage(
		string $threemaId,
		string $filePath,
		?string $thumbnailPath = null,
		?string $caption = null,
		?string $fileName = null,
		?FileRenderingType $renderingType = FileRenderingType::FILE,
		?string $correlationId = null,
		?array $metadata = null
	): SendE2EResult {
		$sendResult = $this->processFileSending([$threemaId], $filePath, new class ($fileName, $renderingType, $correlationId, $metadata) implements ProcessFileSend {
			public function __construct(private readonly ?string $fileName = null, private readonly ?FileRenderingType $renderingType = null, private readonly ?string $correlationId = null, private readonly ?array $metadata = null)
			{
			}

			public function process(
				CryptTool $cryptTool,
				string $privateKey,
				UploadFileResult $uploadResult,
				EncryptResult $encryptionResult,
				FileAnalysisResult $fileAnalysisResult,
				string $receiverPublicKey,
				string $nonce,
				?UploadFileResult $thumbnailUploadResult,
				?string $thumbnailMimeType,
				?string $caption
			): string {
				$message = new FileMessage(
					$fileAnalysisResult->getMimeType(),
					$fileAnalysisResult->getSize(),
					$uploadResult->getBlobId(),
					$encryptionResult->getKey(),
					$thumbnailUploadResult?->getBlobId(),
					$thumbnailMimeType,
					$caption,
					$this->renderingType,
					$this->correlationId,
					$this->metadata,
					$this->fileName ?? $fileAnalysisResult->getFileName()
				);

				return $cryptTool->encryptFileMessage(
					$message,
					$privateKey,
					$receiverPublicKey,
					$nonce,
				);
			}
		}, $caption, $thumbnailPath);

		if (false === array_key_exists($threemaId, $sendResult)) {
			return new SendE2EResult(-1, '');
		}

		return $this->sendE2E($threemaId, $sendResult[$threemaId]->getNonce(), $sendResult[$threemaId]->getBox());
	}

	/**
	 * @param ?string $thumbnailPath
	 * @param ?string $caption
	 * @param ?string $fileName
	 * @param ?FileRenderingType $renderingType
	 * @param ?string $correlationId
	 * @param ?array $metadata
	 * @throws Exception
	 */
	public function sendFileMessages(
		array $threemaIds,
		string $filePath,
		?string $thumbnailPath = null,
		?string $caption = null,
		?string $fileName = null,
		?FileRenderingType $renderingType = null,
		?string $correlationId = null,
		?array $metadata = null
	): SendE2EBulkResult {
		$threemaIds = array_unique($threemaIds);

		$sendResults = $this->processFileSending($threemaIds, $filePath, new class ($fileName, $renderingType, $correlationId, $metadata) implements ProcessFileSend {
			public function __construct(private readonly ?string $fileName = null, private readonly ?FileRenderingType $renderingType = null, private readonly ?string $correlationId = null, private readonly ?array $metadata = null)
			{
			}

			public function process(
				CryptTool $cryptTool,
				string $privateKey,
				UploadFileResult $uploadResult,
				EncryptResult $encryptionResult,
				FileAnalysisResult $fileAnalysisResult,
				string $receiverPublicKey,
				string $nonce,
				?UploadFileResult $thumbnailUploadResult,
				?string $thumbnailMimeType,
				?string $caption
			): string {
				$message = new FileMessage(
					$fileAnalysisResult->getMimeType(),
					$fileAnalysisResult->getSize(),
					$uploadResult->getBlobId(),
					$encryptionResult->getKey(),
					$thumbnailUploadResult?->getBlobId(),
					$thumbnailMimeType,
					$caption,
					$this->renderingType,
					$this->correlationId,
					$this->metadata,
					$this->fileName ?? $fileAnalysisResult->getFileName()
				);

				return $cryptTool->encryptFileMessage(
					$message,
					$privateKey,
					$receiverPublicKey,
					$nonce
				);
			}
		}, $caption, $thumbnailPath);

		$nonces = [];
		$boxes = [];
		$realThreemaIds = [];
		foreach ($sendResults as $threemaId => $result) {
			$realThreemaIds[] = $threemaId;
			$nonces[] = $result->getNonce();
			$boxes[] = $result->getBox();
		}

		return $this->sendE2EBulk($realThreemaIds, $nonces, $boxes);
	}

	/**
	 * Decrypt a message and download the files of the message to the $outputFolder.
	 * CAUTION: If you set `outputFolder` to a directory inside your webroot an attacker could potentially gain control over your server! Ensure that all received files are stored on a separate server or at least outside the webroot!
	 *
	 * @param string $box box as binary string
	 * @param string $nonce nonce as binary string
	 * @param string|null $outputFolder folder for storing the files (outside webroot!)
	 * @param ?Closure $downloadMessage
	 * @throws BadMessageException
	 * @throws DecryptionFailedException
	 * @throws Exception
	 * @throws UnsupportedMessageTypeException
	 * @throws \Exception
	 */
	public function receiveMessage(
		string $threemaId,
		string $messageId,
		string $box,
		string $nonce,
		?string $outputFolder = null,
		?Closure $downloadMessage = null
	): ReceiveMessageResult {
		//fetch the public key
		$receiverPublicKey = $this->connection->fetchPublicKey($threemaId);

		if (!$receiverPublicKey->isSuccess()) {
			throw new Exception('Invalid threema id');
		}

		$message = $this->cryptTool->decryptMessage(
			$box,
			$this->privateKey,
			hex2bin((string)$receiverPublicKey->getPublicKey()),
			$nonce
		);

		if (false === is_object($message)) {
			throw new Exception('Could not encrypt box');
		}

		$receiveResult = new ReceiveMessageResult($messageId, $message);

		try {
			$randomFilename = bin2hex(random_bytes(16));
		} catch (\Exception) {
			throw new Exception('Saving the downloaded file(s) failed!');
		}

		// Just to be safe
		if (mb_strlen($randomFilename, '8bit') !== 32) {
			throw new Exception('Saving the downloaded file(s) failed!');
		}

		if ($message instanceof FileMessage || $message instanceof GroupFileMessage) {
			$savedCandidate = $this->processDownload(
				$message,
				$message->getBlobId(),
				$nonce,
				function (string $data) use ($message) {
					return $this->cryptTool->decryptFile(
						$data,
						hex2bin($message->getEncryptionKey())
					);
				},
				$randomFilename,
				$outputFolder,
				$downloadMessage
			);

			if (null !== $savedCandidate) {
				$filePath = $outputFolder . '/' . $randomFilename;
				$receiveResult->addFile('file', $filePath);

				if (null !== $message->getThumbnailBlobId() && strlen($message->getThumbnailBlobId()) > 0) {
					$ext = '';
					if (strlen($outputFolder) > 0) {
						if ($message->getThumbnailMediaType() == 'image/jpg') {
							$ext = '.jpg';
						} elseif ($message->getThumbnailMediaType() == 'image/jpeg') {
							$ext = '.jpeg';
						} elseif ($message->getThumbnailMediaType() == 'image/png') {
							$ext = '.png';
						}
					}
					$savedThumbnail = $this->processDownload(
						$message,
						$message->getThumbnailBlobId(),
						$nonce,
						function (string $data) use ($message) {
							return $this->cryptTool->decryptFileThumbnail(
								$data,
								hex2bin($message->getEncryptionKey())
							);
						},
						$randomFilename,
						$outputFolder,
						$downloadMessage,
						$ext
					);

					if (null !== $savedThumbnail) {
						$filePath = $outputFolder . '/' . 'thumbnail-' . $randomFilename . $ext;
						$receiveResult->addFile('thumbnail', $filePath);
					}
				}
			}
		} elseif ($message instanceof ImageMessage) {
			$savedCandidate = $this->processDownload(
				$message,
				$message->getBlobId(),
				$message->getNonce(),
				function (string $data, string $nonce) use ($receiverPublicKey) {
					return $this->cryptTool->decryptImage(
						$data,
						hex2bin((string)$receiverPublicKey->getPublicKey()),
						$this->privateKey,
						$nonce
					);
				},
				$randomFilename,
				$outputFolder,
				$downloadMessage
			);

			if (null !== $savedCandidate) {
				$filePath = "{$outputFolder}/{$randomFilename}";
				$receiveResult->addFile('image', $filePath);

				// Extract caption
				$result = $this->getFileAnalysisTool()->analyse($filePath);
				if (null !== $result && is_array($result->getExifTags())) {
					$exifTags = $result->getExifTags();
					foreach (['UserComment', 'Artist'] as $exifTag) {
						if (true === array_key_exists($exifTag, $exifTags)
							&& strlen((string)$exifTags[$exifTag])) {
							$message->setCaption($exifTags[$exifTag]);

							break;
						}
					}
				}
			}
		} elseif ($message instanceof VideoMessage) {
			$savedCandidate = $this->processDownload(
				$message,
				$message->getBlobId(),
				$nonce,
				function (string $data) use ($message) {
					return $this->cryptTool->decryptFile(
						$data,
						hex2bin($message->getEncryptionKey())
					);
				},
				$randomFilename,
				$outputFolder,
				$downloadMessage
			);

			if (null !== $savedCandidate) {
				$filePath = "{$outputFolder}/{$randomFilename}";
				$receiveResult->addFile('video', $filePath);

				if (null !== $message->getThumbnailBlobId() && strlen((string)$message->getThumbnailBlobId()) > 0) {
					$ext = '.jpg';
					$savedThumbnail = $this->processDownload(
						$message,
						$message->getThumbnailBlobId(),
						$nonce,
						function (string $data) use ($message) {
							return $this->cryptTool->decryptFileThumbnail(
								$data,
								hex2bin($message->getEncryptionKey())
							);
						},
						$randomFilename,
						$outputFolder,
						$downloadMessage,
						$ext
					);

					if (null !== $savedThumbnail) {
						$filePath = $outputFolder . '/' . 'thumbnail-' . $randomFilename . $ext;
						$receiveResult->addFile('thumbnail', $filePath);
					}
				}
			}
		} elseif ($message instanceof AudioMessage) {
			$savedCandidate = $this->processDownload(
				$message,
				$message->getBlobId(),
				$nonce,
				function (string $data) use ($message) {
					return $this->cryptTool->decryptFile(
						$data,
						hex2bin($message->getEncryptionKey())
					);
				},
				$randomFilename,
				$outputFolder,
				$downloadMessage
			);

			if (null !== $savedCandidate) {
				$filePath = "{$outputFolder}/{$randomFilename}";
				$receiveResult->addFile('audioFile', $filePath);
			}
		}

		return $receiveResult;
	}

	/**
	 * FIXME: Does this *have to* bulk send?
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendBallotCreateMessages(
		BallotCreateMessage $message,
		array $threemaIds,
		?array $moreOptions = null
	): SendE2EBulkResult {
		$options = $moreOptions ?? [];
		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptBallotCreateMessage(
				$message,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupBallotCreateMessages(
		GroupBallotCreateMessage $message,
		array $threemaIds,
		?array $moreOptions = null
	): SendE2EBulkResult {
		$options = $moreOptions ?? [];
		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptGroupBallotCreateMessage(
				$message,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendGroupBallotVoteMessages(
		GroupBallotVoteMessage $message,
		array $threemaIds,
		?array $moreOptions = null
	): SendE2EBulkResult {
		$options = $moreOptions ?? [];
		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptGroupBallotVoteMessage(
				$message,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}

	/**
	 * @param ?array $moreOptions
	 * @throws Exception
	 */
	public function sendBallotVoteMessages(
		BallotVoteMessage $message,
		array $threemaIds,
		?array $moreOptions = null
	): SendE2EBulkResult {
		$options = $moreOptions ?? [];
		$threemaIds = array_unique($threemaIds);

		// fetch all public keys
		$publicKeys = $this->connection->fetchPublicKeyBulk($threemaIds);
		if (false === $publicKeys->isSuccess() || count($threemaIds)
			!== (is_countable($publicKeys->getPublicKeys()) ? count($publicKeys->getPublicKeys()) : 0)) {
			throw new Exception('Could not fetch all public keys');
		}

		$nonces = [];
		$boxes = [];
		foreach ($threemaIds as $threemaId) {
			$nonce = $this->getCryptTool()->randomNonce();

			$nonces[] = $nonce;
			$boxes[] = $this->cryptTool->encryptBallotVoteMessage(
				$message,
				$this->privateKey,
				$publicKeys->getPublicKey($threemaId),
				$nonce
			);
		}

		return $this->sendE2EBulk($threemaIds, $nonces, $boxes, $options);
	}
}
