<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\MsgApi\PublicKeyStores;

use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\PublicKeyStore;

/**
 * Store the PublicKeys in a PHP file
 */
class PhpFile extends PublicKeyStore
{
	/** PHP code used to prevent unauthorized access */
	private string $fileBlocker = '<?php if (!isset($isMsgApiKeystore) || !$isMsgApiKeystore) die(\'Unauthorized access\');';

	private readonly string $file;

	/** cache of array */
	private array $keystore = [];

	/**
	 * @param string $file path to a read and writable (empty) PHP file
	 * @throws Exception if the file does not exist, is not writable or no valid PHP file
	 */
	public function __construct(string $file)
	{
		if (false === is_writable($file)) {
			throw new Exception('file ' . $file . ' does not exist or is not writable');
		}
		if (pathinfo($file, PATHINFO_EXTENSION) != 'php') {
			throw new Exception('file ' . $file . ' is not a valid PHP file');
		}
		$this->file = $file;
	}

	/**
	 * initiate a php file
	 *
	 * @throws Exception
	 */
	private function initFile(mixed $fileHandle = null): bool
	{
		//check if file does already contain content
		if (filesize($this->file) != 0) {
			return true;
		}

		//manually open file if no file handle is given
		$fileopened = null;
		if (null === $fileHandle) {
			$fileHandle = fopen($this->file, 'r+');
			$fileopened = true;

			//check for success
			if (null === $fileHandle) {
				throw new Exception('could not open file ' . $this->file);
			}
		}

		//create content
		$content = $this->fileBlocker . PHP_EOL;
		$content .= PHP_EOL;
		$content .= '//Threema MsgApi phpfile keystore' . PHP_EOL;
		$content .= '//DO NOT EDIT THIS FILE!' . PHP_EOL;
		$content .= PHP_EOL;

		//write file
		$fwrite = fwrite($fileHandle, $content);
		if (!$fwrite) {
			throw new Exception('could not write to file ' . $this->file);
		}

		//close file if necessary
		if ($fileopened) {
			$fclose = fclose($fileHandle);
			if (!$fclose) {
				throw new Exception('error while processing file ' . $this->file);
			}
		}

		return true;
	}

	/**
	 * return null if the public key not found in the store
	 */
	public function findPublicKey(string $threemaId): ?string
	{
		$threemaId = strtoupper($threemaId);
		$publicKey = null;

		//Pre-check loaded array
		if (array_key_exists($threemaId, $this->keystore)) {
			return $this->keystore[$threemaId];
		}

		//Parse file
		$keystore = [];
		$isMsgApiKeystore = true;
		require $this->file;

		//Update cache
		$this->keystore = $keystore;

		//Check file content
		if (array_key_exists($threemaId, $keystore)) {
			$publicKey = $keystore[$threemaId];
		}

		return $publicKey;
	}

	/**
	 * save a public key
	 *
	 * @throws Exception
	 */
	public function savePublicKey(string $threemaId, string $publicKey): bool
	{
		//create file if needed
		$this->initFile();

		//check for key
		if (array_key_exists($threemaId, $this->keystore)) {
			return false;
		}

		//add key
		$this->keystore[$threemaId] = $publicKey;
		$content = '$keystore[\'' . $threemaId . '\'] = \'' . $publicKey . '\';' . PHP_EOL;

		//write content
		$fileadd = file_put_contents($this->file, $content, FILE_APPEND);
		if (!$fileadd) {
			throw new Exception('could not write to file ' . $this->file);
		}

		return true;
	}

	/**
	 * Initialize a new PhpFile Public Key Store
	 * @param string $path the file will be created it does not exist
	 * @throws Exception
	 */
	public static function create(string $path): PhpFile
	{
		if (false === file_exists($path)) {
			//touch
			touch($path);
		}

		return new PhpFile($path);
	}
}
