<?php
/**
 * PHP version 8.1
 *
 * @category  Class
 * @author    Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see       https://gateway.threema.ch/en/developer
 */

/**
 * Threema Gateway Php SDK
 *
 * This SDK allows for preparing, sending and receiving of Threema Messages via Threema Gateway.
 *
 * Generated by: https://gateway.threema.ch/en/developer
 * MsgApi SDK Generator version: 0.1.0
 */

namespace Threema\MsgApi\Tools;

use Stringable;
use Threema\MsgApi\Commands\Results\UploadFileResult;
use Threema\MsgApi\Constants;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Core\KeyPair;
use Threema\MsgApi\Exceptions\BadMessageException;
use Threema\MsgApi\Exceptions\DecryptionFailedException;
use Threema\MsgApi\Exceptions\UnsupportedMessageTypeException;
use Threema\MsgApi\Messages;
use Threema\MsgApi\Types;
use Threema\MsgApi\Types\GroupId;

/**
 * Interface CryptTool
 * Contains static methods to do various Threema cryptography related tasks.
 */
abstract class CryptTool implements Stringable
{
	public const TYPE_SODIUM = 'sodium';
	public const TYPE_SALT = 'salt';
	public const FILE_NONCE = "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01";
	public const FILE_THUMBNAIL_NONCE = "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02";
	public const EMAIL_HMAC_KEY = "\x30\xa5\x50\x0f\xed\x97\x01\xfa\x6d\xef\xdb\x61\x08\x41\x90\x0f\xeb\xb8\xe4\x30\x88\x1f\x7a\xd8\x16\x82\x62\x64\xec\x09\xba\xd7";
	public const PHONENO_HMAC_KEY = "\x85\xad\xf8\x22\x69\x53\xf3\xd9\x6c\xfd\x5d\x09\xbf\x29\x55\x5e\xb9\x55\xfc\xd8\xaa\x5e\xc4\xf9\xfc\xd8\x69\xe2\x58\x37\x07\x23";

	private static ?CryptTool $instance = null;

	protected function __construct()
	{
	}

	public function __toString(): string
	{
		return 'CryptTool ' . $this->getName();
	}

	/**
	 * determine random amount of PKCS7 padding
	 */
	private function generatePadBytes(): int
	{
		$padbytes = 0;
		while ($padbytes < 1 || $padbytes > 255) {
			$padbytes = ord($this->createRandom(1));
		}

		return $padbytes;
	}

	/**
	 * make a box
	 *
	 * @return string encrypted box
	 */
	abstract protected function makeBox(string $data, string $nonce, string $senderPrivateKey, string $recipientPublicKey): string;

	abstract protected function createRandom(int $size);

	/**
	 * Manually set or clear the cryptTool instance
	 * @param ?CryptTool $tool
	 */
	public static function setInstance(?CryptTool $tool = null)
	{
		self::$instance = $tool;
	}

	/**
	 * Prior libsodium
	 */
	public static function getInstance(): ?CryptTool
	{
		if (null === self::$instance) {
			foreach ([
				fn () => self::createInstance(self::TYPE_SODIUM),
				fn () => self::createInstance(self::TYPE_SALT)] as $instanceGenerator) {
				$i = $instanceGenerator->__invoke();
				if (null !== $i) {
					self::setInstance($i);

					break;
				}
			}
		}

		return self::$instance;
	}

	/**
	 * @return CryptTool|CryptToolSalt|CryptToolSodiumIntegrated|null null on unknown type
	 */
	public static function createInstance(string $type): CryptTool|CryptToolSalt|null
	{
		switch ($type) {
			case self::TYPE_SODIUM:
				$instance = new CryptToolSodiumIntegrated();
				if (true === $instance->isSupported()) {
					return $instance;
				}

				return $instance->isSupported() ? $instance : null;
			case self::TYPE_SALT:
				$instance = new CryptToolSalt();

				return $instance->isSupported() ? $instance : null;
			default:
				return null;
		}
	}

	/*
	 * =====================
	 * Message types
	 * =====================
	 */

	/**
	 * @return string Result of the Text encryption
	 */
	public function encryptTextMessage(
		Messages\TextMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getText();

		return $this->padAndMakeBox(
			Messages\TextMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Location encryption
	 */
	public function encryptLocationMessage(
		Messages\LocationMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= sprintf('%f,%f', $message->getLatitude(), $message->getLongitude());
		if (!is_null($message->getAccuracy())) {
			$content .= sprintf(',%f', $message->getAccuracy());
		}

		if (!empty($message->getPoiName())) {
			$content .= "\n" . $message->getPoiName();
		}

		if (!empty($message->getPoiAddress())) {
			$content .= "\n" . str_replace("\n", '\\n', $message->getPoiAddress());
		}

		return $this->padAndMakeBox(
			Messages\LocationMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the File encryption
	 */
	public function encryptFileMessage(
		Messages\FileMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\FileMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Ballot encryption
	 */
	public function encryptBallotCreateMessage(
		Messages\BallotCreateMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getBallotId()->getBallotId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\BallotCreateMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Ballot encryption
	 */
	public function encryptBallotVoteMessage(
		Messages\BallotVoteMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getCreator();
		$content .= $message->getBallotId()->getBallotId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\BallotVoteMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Delivery encryption
	 */
	public function encryptDeliveryReceipt(
		Messages\DeliveryReceipt $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->jsonSerialize();

		return $this->padAndMakeBox(
			Messages\DeliveryReceipt::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupCreateMessage(
		Messages\GroupCreateMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupCreateMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupRenameMessage(
		Messages\GroupRenameMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupRenameMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupLeaveMessage(
		Messages\GroupLeaveMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupLeaveMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupSetPhoto(
		Messages\GroupSetPhoto $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupSetPhoto::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupDeletePhoto(
		Messages\GroupDeletePhoto $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupDeletePhoto::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupRequestSyncMessage(
		Messages\GroupRequestSyncMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupId();

		return $this->padAndMakeBox(
			Messages\GroupRequestSyncMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupTextMessage(
		Messages\GroupTextMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= $message->getText();

		return $this->padAndMakeBox(
			Messages\GroupTextMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupLocationMessage(
		Messages\GroupLocationMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= sprintf('%f,%f', $message->getLatitude(), $message->getLongitude());
		if (!is_null($message->getAccuracy())) {
			$content .= sprintf(',%f', $message->getAccuracy());
		}

		if (!empty($message->getPoiName())) {
			$content .= "\n" . $message->getPoiName();
		}

		if (!empty($message->getPoiAddress())) {
			$content .= "\n" . str_replace("\n", '\\n', $message->getPoiAddress());
		}

		return $this->padAndMakeBox(
			Messages\GroupLocationMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupFileMessage(
		Messages\GroupFileMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupFileMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupBallotCreateMessage(
		Messages\GroupBallotCreateMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= $message->getBallotId()->getBallotId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupBallotCreateMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Group encryption
	 */
	public function encryptGroupBallotVoteMessage(
		Messages\GroupBallotVoteMessage $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= $message->getCreator();
		$content .= $message->getBallotId()->getBallotId();
		$content .= json_encode($message->jsonSerialize());

		return $this->padAndMakeBox(
			Messages\GroupBallotVoteMessage::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @return string Result of the Delivery encryption
	 */
	public function encryptGroupDeliveryReceipt(
		Messages\GroupDeliveryReceipt $message,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$content = '';
		$content .= $message->getGroupId()->getGroupCreator();
		$content .= $message->getGroupId()->getGroupId();
		$content .= $message->jsonSerialize();

		return $this->padAndMakeBox(
			Messages\GroupDeliveryReceipt::TYPE_CODE,
			$content,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/*
	 * =========================
	 * Legacy types (deprecated)
	 * =========================
	 */

	/**
	 * @param UploadFileResult $uploadFileResult the result of the upload
	 * @param EncryptResult $encryptResult the result of the image encryption
	 * @param string $senderPrivateKey the private key of the sending ID (as binary)
	 * @param string $recipientPublicKey the public key of the receiving ID (as binary)
	 * @param string $nonce the nonce to be used for the encryption (usually 24 random bytes)
	 */
	public function encryptImageMessage(
		UploadFileResult $uploadFileResult,
		EncryptResult $encryptResult,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$message = hex2bin((string)$uploadFileResult->getBlobId());
		$message .= pack('V', $encryptResult->getSize());
		$message .= $encryptResult->getNonce();

		return $this->padAndMakeBox(
			Messages\ImageMessage::TYPE_CODE,
			$message,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @param int $duration duration of the video (in seconds)
	 * @param UploadFileResult $uploadFileResult the result of the video file upload
	 * @param UploadFileResult $uploadThumbnailResult the result of the video thumbnail file upload
	 * @param EncryptResult $encryptResult the result of the video encryption
	 * @param EncryptResult $encryptThumbnailResult the result of the thumbnail encryption
	 * @param string $senderPrivateKey the private key of the sending ID (as binary)
	 * @param string $recipientPublicKey the public key of the receiving ID (as binary)
	 * @param string $nonce the nonce to be used for the encryption (usually 24 random bytes)
	 */
	public function encryptVideoMessage(
		int $duration,
		UploadFileResult $uploadFileResult,
		UploadFileResult $uploadThumbnailResult,
		EncryptResult $encryptResult,
		EncryptResult $encryptThumbnailResult,
		string $senderPrivateKey,
		string $recipientPublicKey,
		string $nonce
	): string {
		$message = pack('s', $duration);
		$message .= hex2bin((string)$uploadFileResult->getBlobId());
		$message .= pack('V', $encryptResult->getSize());
		$message .= hex2bin((string)$uploadThumbnailResult->getBlobId());
		$message .= pack('V', $encryptThumbnailResult->getSize());
		$message .= $encryptResult->getKey();

		return $this->padAndMakeBox(
			Messages\VideoMessage::TYPE_CODE,
			$message,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @param int $duration duration of the audio (in seconds)
	 * @param UploadFileResult $uploadFileResult the result of the audio file upload
	 * @param EncryptResult $encryptResult the result of the video encryption
	 * @param string $senderPrivateKey the private key of the sending ID (as binary)
	 * @param string $recipientPublicKey the public key of the receiving ID (as binary)
	 * @param string $nonce the nonce to be used for the encryption (usually 24 random bytes)
	 */
	public function encryptAudioMessage(
		$duration,
		UploadFileResult $uploadFileResult,
		EncryptResult $encryptResult,
		$senderPrivateKey,
		$recipientPublicKey,
		$nonce
	): string {
		$message = pack('s', $duration);
		$message .= hex2bin((string)$uploadFileResult->getBlobId());
		$message .= pack('V', $encryptResult->getSize());
		$message .= $encryptResult->getKey();

		return $this->padAndMakeBox(
			Messages\AudioMessage::TYPE_CODE,
			$message,
			$nonce,
			$senderPrivateKey,
			$recipientPublicKey
		);
	}

	/**
	 * @param string $typeCode (binary)
	 * @param string $messageContent (binary)
	 * @param string $nonce (binary)
	 * @param string $senderPrivateKey (binary)
	 * @param string $recipientPublicKey (binary)
	 */
	final public function padAndMakeBox(string $typeCode, string $messageContent, string $nonce, string $senderPrivateKey, string $recipientPublicKey): string
	{
		// Append type code
		$message = pack('C', $typeCode);

		// Append content
		$message .= $messageContent;

		// determine random amount of PKCS7 padding
		$padbytes = $this->generatePadBytes();

		// Ensure padded_data amounts to at least 32 bytes as per CSP advisory
		$msgLen = strlen($message);
		if ($msgLen < 32) {
			$padbytes = 32 - $msgLen;
		}

		// append padding
		$message .= str_repeat(chr($padbytes), $padbytes);

		return $this->makeBox($message, $nonce, $senderPrivateKey, $recipientPublicKey);
	}

	/**
	 * make a secret box
	 */
	abstract public function makeSecretBox(string $data, string $nonce, string $key): mixed;

	/**
	 * decrypt a box
	 *
	 * @param string $box as binary
	 * @param string $recipientPrivateKey as binary
	 * @param string $senderPublicKey as binary
	 * @param string $nonce as binary
	 */
	abstract public function openBox(string $box, string $recipientPrivateKey, string $senderPublicKey, string $nonce): string;

	/**
	 * decrypt a secret box
	 *
	 * @param string $box as binary
	 * @param string $nonce as binary
	 * @param string $key as binary
	 * @return string as binary
	 */
	abstract public function openSecretBox(string $box, string $nonce, string $key): string;

	/**
	 * Decrypts a ThreemaMessage
	 *
	 * @throws DecryptionFailedException
	 * @throws Exception
	 * @throws UnsupportedMessageTypeException
	 * @throws BadMessageException
	 * @throws \Exception
	 * @return Messages\ThreemaMessage the decrypted message
	 */
	public function decryptMessage(string $box, string $recipientPrivateKey, string $senderPublicKey, string $nonce): Messages\ThreemaMessage
	{
		$data = $this->openBox($box, $recipientPrivateKey, $senderPublicKey, $nonce);

		if (null === $data || strlen($data) == 0) {
			throw new DecryptionFailedException();
		}

		// remove padding
		$padbytes = ord($data[strlen($data) - 1]);
		$realDataLength = strlen($data) - $padbytes;
		if ($realDataLength < 1) {
			throw new BadMessageException();
		}
		$data = substr($data, 0, $realDataLength);

		// first byte of data is type
		$type = ord($data[0]);

		$pos = 1;
		$piece = function ($length) use (&$pos, $data): string {
			$d = substr($data, $pos, $length);
			$pos += $length;

			return $d;
		};

//		var_dump($data);

		switch ($type) {
			case Messages\ImageMessage::TYPE_CODE:
				// Image Message
				if (1 + Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN + Constants::IMAGE_NONCE_LEN != $realDataLength) {
					throw new BadMessageException();
				}

				$blobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$length = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$nonce = $piece->__invoke(Constants::IMAGE_NONCE_LEN);

				return new Messages\ImageMessage(bin2hex((string)$blobId), current(unpack('V', (string)$length)), $nonce);
			case Messages\VideoMessage::TYPE_CODE:
				// Video Message
				if (1 + 2
					+ Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN
					+ Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN
					+ Constants::SYMMETRIC_KEY_LEN != $realDataLength) {
					throw new BadMessageException();
				}

				$duration = $piece->__invoke(2);
				$blobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$blobLength = $piece->__invoke(Constants::FILE_SIZE_LEN);

				$thumbnailBlobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$thumbnailLength = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$key = $piece->__invoke(Constants::SYMMETRIC_KEY_LEN);

				return new Messages\VideoMessage(
					bin2hex((string)$blobId),
					current(unpack('V', ((string)$blobLength))),
					bin2hex((string)$thumbnailBlobId),
					current(unpack('V', ((string)$thumbnailLength))),
					bin2hex((string)$key),
					current(unpack('s', (string)$duration))
				);

			case Messages\AudioMessage::TYPE_CODE:
				// Audio Message
				if (1 + 2
					+ Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN
					+ Constants::SYMMETRIC_KEY_LEN != $realDataLength) {
					throw new BadMessageException();
				}

				$duration = $piece->__invoke(2);
				$blobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$blobLength = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$key = $piece->__invoke(Constants::SYMMETRIC_KEY_LEN);

				return new Messages\AudioMessage(
					bin2hex((string)$blobId),
					current(unpack('V', ((string)$blobLength))),
					bin2hex((string)$key),
					current(unpack('s', (string)$duration))
				);
			case Messages\GroupImageMessage::TYPE_CODE:
				if ($realDataLength !== (1 + Constants::IDENTITY_LEN + Constants::GROUP_ID_LEN + Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN + Constants::SYMMETRIC_KEY_LEN)) {
					throw new UnsupportedMessageTypeException('invalid group image message length');
				}
				$groupId = new GroupId(
					$piece->__invoke(Constants::IDENTITY_LEN),
					$piece->__invoke(Constants::GROUP_ID_LEN)
				);

				$blobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$blobLength = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$key = $piece->__invoke(Constants::SYMMETRIC_KEY_LEN);

				return new Messages\GroupImageMessage(
					$groupId,
					bin2hex((string)$blobId),
					current(unpack('V', (string)$blobLength)),
					bin2hex((string)$key)
				);

			case Messages\GroupAudioMessage::TYPE_CODE:
				// Audio Message
				if (1
					+ Constants::IDENTITY_LEN + Constants::GROUP_ID_LEN
					+ 2
					+ Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN
					+ Constants::SYMMETRIC_KEY_LEN != $realDataLength) {
					throw new BadMessageException();
				}
				$groupId = new GroupId(
					$piece->__invoke(Constants::IDENTITY_LEN),
					$piece->__invoke(Constants::GROUP_ID_LEN)
				);
				$duration = $piece->__invoke(2);
				$blobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$blobLength = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$key = $piece->__invoke(Constants::SYMMETRIC_KEY_LEN);

				return new Messages\GroupAudioMessage(
					$groupId,
					bin2hex((string)$blobId),
					current(unpack('V', ((string)$blobLength))),
					bin2hex((string)$key),
					current(unpack('s', (string)$duration))
				);
			case Messages\GroupVideoMessage::TYPE_CODE:
				// Audio Message
				if (1
					+ Constants::IDENTITY_LEN + Constants::GROUP_ID_LEN
					// Duration
					+ 2
					// Video
					+ Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN
					// Thumbnail
					+ Constants::BLOB_ID_LEN + Constants::FILE_SIZE_LEN
					+ Constants::SYMMETRIC_KEY_LEN != $realDataLength) {
					throw new BadMessageException();
				}
				$groupId = new GroupId(
					$piece->__invoke(Constants::IDENTITY_LEN),
					$piece->__invoke(Constants::GROUP_ID_LEN)
				);
				$duration = $piece->__invoke(2);
				$blobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$blobLength = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$thumbnailBlobId = $piece->__invoke(Constants::BLOB_ID_LEN);
				$thumbnailBlobLength = $piece->__invoke(Constants::FILE_SIZE_LEN);
				$key = $piece->__invoke(Constants::SYMMETRIC_KEY_LEN);

				return new Messages\GroupVideoMessage(
					$groupId,
					bin2hex((string)$blobId),
					current(unpack('V', ((string)$blobLength))),
					bin2hex((string)$thumbnailBlobId),
					current(unpack('V', ((string)$thumbnailBlobLength))),
					bin2hex((string)$key),
					current(unpack('s', (string)$duration))
				);
			default:
		}

		return match ($type) {
			Messages\TextMessage::TYPE_CODE => Messages\TextMessage::fromString($data, $realDataLength, $piece),
			Messages\LocationMessage::TYPE_CODE => Messages\LocationMessage::fromString($data, $realDataLength, $piece),
			Messages\FileMessage::TYPE_CODE => Messages\FileMessage::fromString($data, $realDataLength, $piece),
			Messages\BallotCreateMessage::TYPE_CODE => Messages\BallotCreateMessage::fromString($data, $realDataLength, $piece),
			Messages\BallotVoteMessage::TYPE_CODE => Messages\BallotVoteMessage::fromString($data, $realDataLength, $piece),
			Messages\DeliveryReceipt::TYPE_CODE => Messages\DeliveryReceipt::fromString($data, $realDataLength, $piece),
			Messages\GroupCreateMessage::TYPE_CODE => Messages\GroupCreateMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupRenameMessage::TYPE_CODE => Messages\GroupRenameMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupLeaveMessage::TYPE_CODE => Messages\GroupLeaveMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupSetPhoto::TYPE_CODE => Messages\GroupSetPhoto::fromString($data, $realDataLength, $piece),
			Messages\GroupDeletePhoto::TYPE_CODE => Messages\GroupDeletePhoto::fromString($data, $realDataLength, $piece),
			Messages\GroupRequestSyncMessage::TYPE_CODE => Messages\GroupRequestSyncMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupTextMessage::TYPE_CODE => Messages\GroupTextMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupLocationMessage::TYPE_CODE => Messages\GroupLocationMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupFileMessage::TYPE_CODE => Messages\GroupFileMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupBallotCreateMessage::TYPE_CODE => Messages\GroupBallotCreateMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupBallotVoteMessage::TYPE_CODE => Messages\GroupBallotVoteMessage::fromString($data, $realDataLength, $piece),
			Messages\GroupDeliveryReceipt::TYPE_CODE => Messages\GroupDeliveryReceipt::fromString($data, $realDataLength, $piece),
			default => throw new UnsupportedMessageTypeException(),
		};
	}

	/**
	 * Generate a new key pair.
	 *
	 * @return KeyPair the new key pair
	 */
	abstract public function generateKeyPair(): KeyPair;

	/**
	 * Hashes an email address for identity lookup.
	 *
	 * @param string $email the email address
	 * @return string the email hash (hex)
	 */
	public function hashEmail(string $email): string
	{
		$emailClean = strtolower(trim($email));

		return hash_hmac('sha256', $emailClean, self::EMAIL_HMAC_KEY);
	}

	/**
	 * Hashes a phone number for identity lookup.
	 *
	 * @param string $phoneNo the phone number (in E.164 format, no leading +)
	 * @return string the phone number hash (hex)
	 */
	public function hashPhoneNo(string $phoneNo): string
	{
		$phoneNoClean = preg_replace('/[^0-9]/', '', $phoneNo);

		return hash_hmac('sha256', $phoneNoClean, self::PHONENO_HMAC_KEY);
	}

	/**
	 * Generate a random nonce.
	 *
	 * @return string random nonce
	 */
	public function randomNonce(): string
	{
		return $this->createRandom(SODIUM_CRYPTO_BOX_NONCEBYTES);
	}

	/**
	 * Generate a symmetric key
	 */
	public function symmetricKey(): mixed
	{
		return $this->createRandom(32);
	}

	/**
	 * Derive the public key
	 *
	 * @param string $privateKey as binary
	 * @return string as binary
	 */
	abstract public function derivePublicKey(string $privateKey): string;

	/**
	 * Check if implementation supported
	 */
	abstract public function isSupported(): bool;

	/**
	 * Validate crypt tool
	 *
	 * @throws Exception
	 */
	abstract public function validate(): bool;

	public function encryptFile(string $data): EncryptResult
	{
		$key = $this->symmetricKey();
		$box = $this->makeSecretBox($data, self::FILE_NONCE, $key);

		return new EncryptResult($box, $key, self::FILE_NONCE, strlen((string)$box));
	}

	/**
	 * @param string $key as binary
	 */
	public function decryptFile(string $data, string $key): ?string
	{
		$result = $this->openSecretBox($data, self::FILE_NONCE, $key);

		return false === $result ? null : $result;
	}

	public function encryptFileThumbnail(string $data, string $key): EncryptResult
	{
		$box = $this->makeSecretBox($data, self::FILE_THUMBNAIL_NONCE, $key);

		return new EncryptResult($box, $key, self::FILE_THUMBNAIL_NONCE, strlen((string)$box));
	}

	public function decryptFileThumbnail(string $data, string $key): ?string
	{
		$result = $this->openSecretBox($data, self::FILE_THUMBNAIL_NONCE, $key);

		return false === $result ? null : $result;
	}

	/**
	 * @param string $privateKey as binary
	 * @param string $publicKey as binary
	 */
	public function encryptImage(string $imageData, string $privateKey, string $publicKey): EncryptResult
	{
		$nonce = $this->randomNonce();

		$box = $this->makeBox(
			$imageData,
			$nonce,
			$privateKey,
			$publicKey
		);

		return new EncryptResult($box, '', $nonce, strlen($box));
	}

	/**
	 * @param string $publicKey as binary
	 * @param string $privateKey as binary
	 * @param string $nonce as binary
	 */
	public function decryptImage(string $data, string $publicKey, string $privateKey, string $nonce): string
	{
		return $this->openBox(
			$data,
			$privateKey,
			$publicKey,
			$nonce
		);
	}

	/**
	 * Name of the CryptTool
	 */
	abstract public function getName(): string;

	/**
	 * Description of the CryptTool
	 */
	abstract public function getDescription(): string;
}
