<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\MsgApi\Tools;

use FieldElement;
use Salt;
use SaltException;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Core\KeyPair;
use Threema\MsgApi\Exceptions\DecryptionFailedException;

/**
 * Contains static methods to do various Threema cryptography related tasks.
 */
class CryptToolSalt extends CryptTool
{
	/**
	 * @throws SaltException
	 * @return string encrypted box
	 */
	protected function makeBox(string $data, string $nonce, string $senderPrivateKey, string $recipientPublicKey): string
	{
		return Salt::box($data, $senderPrivateKey, $recipientPublicKey, $nonce)
			->slice(16)->toString();
	}

	/**
	 * @param int $size
	 * @throws SaltException
	 * @return string
	 */
	protected function createRandom($size)
	{
		return Salt::randombytes($size);
	}

	/**
	 * @param string $nonce
	 * @param string $key
	 * @param mixed $data
	 * @throws SaltException
	 * @return string encrypted secret box
	 */
	public function makeSecretBox($data, $nonce, $key): string
	{
		return Salt::secretbox($data, $nonce, $key)
			->slice(16)->toString();
	}

	/**
	 * @throws DecryptionFailedException
	 */
	public function openBox(string $box, string $recipientPrivateKey, string $senderPublicKey, string $nonce): string
	{
		$boxPad = str_repeat("\x00", 16) . $box;

		try {
			$data = Salt::box_open($boxPad, $recipientPrivateKey, $senderPublicKey, $nonce);
		} catch (SaltException) {
			$data = null;
		}

		if ($data) {
			return substr($data->toString(), 32);
		}

		throw new DecryptionFailedException();
	}

	/**
	 * decrypt a secret box
	 *
	 * @param string $box as binary
	 * @param string $nonce as binary
	 * @param string $key as binary
	 * @throws DecryptionFailedException|SaltException
	 * @return string as binary
	 */
	public function openSecretBox(string $box, string $nonce, string $key): string
	{
		$boxPad = str_repeat("\x00", 16) . $box;
		$data = Salt::secretbox_open($boxPad, $nonce, $key);

		if ($data) {
			return substr($data->toString(), 32);
		}

		throw new DecryptionFailedException();
	}

	/**
	 * Generate a new key pair.
	 *
	 * @return KeyPair the new key pair
	 */
	final public function generateKeyPair(): KeyPair
	{
		list($privateKeyObject, $publicKeyObject) = Salt::box_keypair();

		return new KeyPair($privateKeyObject->toString(), $publicKeyObject->toString());
	}

	/**
	 * Derive the public key
	 *
	 * @param string $privateKey in binary
	 * @return string public key as binary
	 */
	final public function derivePublicKey(string $privateKey): string
	{
		$privateKeyElement = FieldElement::fromString($privateKey);

		return Salt::instance()->crypto_scalarmult_base($privateKeyElement)->toString();
	}

	/**
	 * @param mixed $imageData
	 * @param mixed $recipientPublicKey
	 * @param mixed $senderPrivateKey
	 * @throws Exception|SaltException
	 * @return EncryptResult
	 */
	public function encryptImageData($imageData, $recipientPublicKey, $senderPrivateKey)
	{
		$message = Salt::decodeInput($imageData);
		$nonce = $this->randomNonce();
		$salt = Salt::instance();

		//secret key
		$key = $salt->scalarmult($senderPrivateKey, $recipientPublicKey);
		$data = $salt->encrypt(
			$message,
			$message->getSize(),
			Salt::decodeInput($nonce),
			$key
		);

		if (false === $data) {
			throw new Exception('encryption failed');
		}

		return new EncryptResult($data->toString(), $senderPrivateKey, $nonce, strlen($data->toString()));
	}

	/**
	 * Check if implementation supported
	 */
	public function isSupported(): bool
	{
		return class_exists(Salt::class);
	}

	/**
	 * Validate crypt tool
	 *
	 * @throws Exception
	 */
	public function validate(): bool
	{
		if (false === $this->isSupported()) {
			throw new Exception('SALT implementation not supported');
		}

		if (PHP_INT_SIZE < 8) {
			throw new Exception('Pure PHP Crypto implementation requires 64Bit PHP. Please install the libsodium PHP extension.');
		}

		return true;
	}

	public function getName(): string
	{
		return Salt::class;
	}

	/**
	 * Description of the CryptTool
	 */
	public function getDescription(): string
	{
		return 'Pure PHP implementation, please try to install and use the libsodium PHP extension for a better performance';
	}
}
