<?php
/**
 * @author Threema Gateway team
 * @copyright Copyright (c) 2023 Threema GmbH
 * @see https://gateway.threema.ch/en/developer
 */

namespace Threema\MsgApi\Tools;

use function random_bytes;

use SodiumException;
use Threema\MsgApi\Core\Exception;
use Threema\MsgApi\Core\KeyPair;

/**
 * Contains static methods to do various Threema cryptography related tasks.
 * Support PHP (7.2+)-integrated sodium
 */
class CryptToolSodiumIntegrated extends CryptTool
{
	/**
	 * @throws SodiumException
	 * @return string encrypted box
	 */
	protected function makeBox(string $data, string $nonce, string $senderPrivateKey, string $recipientPublicKey): string
	{
		/** @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection */
		$kp = sodium_crypto_box_keypair_from_secretkey_and_publickey($senderPrivateKey, $recipientPublicKey);

		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return sodium_crypto_box($data, $nonce, $kp);
	}

	/**
	 * @param int $size
	 * @throws \Exception
	 * @return string
	 */
	protected function createRandom($size)
	{
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return random_bytes($size);
	}

	/**
	 * make a secret box
	 *
	 * @param mixed $data
	 * @param mixed $nonce
	 * @param mixed $key
	 * @throws SodiumException
	 * @return mixed
	 */
	public function makeSecretBox($data, $nonce, $key): string
	{
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return sodium_crypto_secretbox((string)$data, (string)$nonce, (string)$key);
	}

	/**
	 * @throws SodiumException
	 */
	public function openBox(string $box, string $recipientPrivateKey, string $senderPublicKey, string $nonce): string
	{
		/** @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection */
		$kp = sodium_crypto_box_keypair_from_secretkey_and_publickey($recipientPrivateKey, $senderPublicKey);
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return sodium_crypto_box_open($box, $nonce, $kp);
	}

	/**
	 * decrypt a secret box
	 *
	 * @param string $box as binary
	 * @param string $nonce as binary
	 * @param string $key as binary
	 * @throws SodiumException
	 * @return string as binary
	 */
	public function openSecretBox(string $box, string $nonce, string $key): string
	{
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return sodium_crypto_secretbox_open($box, $nonce, $key);
	}

	/**
	 * Generate a new key pair.
	 *
	 * @throws SodiumException
	 * @return KeyPair the new key pair
	 */
	final public function generateKeyPair(): KeyPair
	{
		/** @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection */
		$kp = sodium_crypto_box_keypair();
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return new KeyPair(sodium_crypto_box_secretkey($kp), sodium_crypto_box_publickey($kp));
	}

	/**
	 * Derive the public key
	 *
	 * @param string $privateKey in binary
	 * @throws SodiumException
	 * @return string public key as binary
	 */
	final public function derivePublicKey(string $privateKey): string
	{
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return sodium_crypto_box_publickey_from_secretkey($privateKey);
	}

	/**
	 * Check if implementation supported
	 */
	public function isSupported(): bool
	{
		return true === extension_loaded('sodium')
			&& true === function_exists('sodium_crypto_box_open');
	}

	/**
	 * Validate crypt tool
	 *
	 * @throws Exception
	 */
	public function validate(): bool
	{
		if (false === $this->isSupported()) {
			throw new Exception('Sodium implementation not supported');
		}

		return true;
	}

	public function getName(): string
	{
		return 'sodium';
	}

	/**
	 * Description of the CryptTool
	 */
	public function getDescription(): string
	{
		// @noinspection PhpUndefinedNamespaceInspection @noinspection PhpUndefinedFunctionInspection
		return 'PHP Sodium implementation';
	}
}
